# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/event_service.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import uri_pb2 as farm__ng_dot_core_dot_uri__pb2
from farm_ng.core import event_pb2 as farm__ng_dot_core_dot_event__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n farm_ng/core/event_service.proto\x12\x12\x66\x61rm_ng.core.proto\x1a\x16\x66\x61rm_ng/core/uri.proto\x1a\x18\x66\x61rm_ng/core/event.proto\"I\n\x10SubscribeRequest\x12$\n\x03uri\x18\x01 \x01(\x0b\x32\x17.farm_ng.core.proto.Uri\x12\x0f\n\x07\x65very_n\x18\x02 \x01(\r\"K\n\x0eSubscribeReply\x12(\n\x05\x65vent\x18\x01 \x01(\x0b\x32\x19.farm_ng.core.proto.Event\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"P\n\x13RequestReplyRequest\x12(\n\x05\x65vent\x18\x01 \x01(\x0b\x32\x19.farm_ng.core.proto.Event\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"N\n\x11RequestReplyReply\x12(\n\x05\x65vent\x18\x01 \x01(\x0b\x32\x19.farm_ng.core.proto.Event\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"~\n\x0cRequestReply\x12\x38\n\x07request\x18\x01 \x01(\x0b\x32\'.farm_ng.core.proto.RequestReplyRequest\x12\x34\n\x05reply\x18\x02 \x01(\x0b\x32%.farm_ng.core.proto.RequestReplyReply\"\x11\n\x0fListUrisRequest\"6\n\rListUrisReply\x12%\n\x04uris\x18\x01 \x03(\x0b\x32\x17.farm_ng.core.proto.Uri\"\xe0\x02\n\x12\x45ventServiceConfig\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\x12\x0c\n\x04host\x18\x03 \x01(\t\x12;\n\rsubscriptions\x18\x04 \x03(\x0b\x32$.farm_ng.core.proto.SubscribeRequest\x12%\n\x04uris\x18\x05 \x03(\x0b\x32\x17.farm_ng.core.proto.Uri\x12\x0c\n\x04\x61rgs\x18\x06 \x03(\t\x12\x42\n\tlog_level\x18\x07 \x01(\x0e\x32/.farm_ng.core.proto.EventServiceConfig.LogLevel\"j\n\x08LogLevel\x12\n\n\x06NOTSET\x10\x00\x12\x0c\n\x08\x43RITICAL\x10\x32\x12\t\n\x05\x46\x41TAL\x10\x32\x12\t\n\x05\x45RROR\x10(\x12\x0b\n\x07WARNING\x10\x1e\x12\x08\n\x04WARN\x10\x1e\x12\x08\n\x04INFO\x10\x14\x12\t\n\x05\x44\x45\x42UG\x10\n\x1a\x02\x10\x01\"Q\n\x16\x45ventServiceConfigList\x12\x37\n\x07\x63onfigs\x18\x01 \x03(\x0b\x32&.farm_ng.core.proto.EventServiceConfig2\xa1\x02\n\x0c\x45ventService\x12T\n\x08listUris\x12#.farm_ng.core.proto.ListUrisRequest\x1a!.farm_ng.core.proto.ListUrisReply\"\x00\x12Y\n\tsubscribe\x12$.farm_ng.core.proto.SubscribeRequest\x1a\".farm_ng.core.proto.SubscribeReply\"\x00\x30\x01\x12`\n\x0crequestReply\x12\'.farm_ng.core.proto.RequestReplyRequest\x1a%.farm_ng.core.proto.RequestReplyReply\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.event_service_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_EVENTSERVICECONFIG_LOGLEVEL']._loaded_options = None
  _globals['_EVENTSERVICECONFIG_LOGLEVEL']._serialized_options = b'\020\001'
  _globals['_SUBSCRIBEREQUEST']._serialized_start=106
  _globals['_SUBSCRIBEREQUEST']._serialized_end=179
  _globals['_SUBSCRIBEREPLY']._serialized_start=181
  _globals['_SUBSCRIBEREPLY']._serialized_end=256
  _globals['_REQUESTREPLYREQUEST']._serialized_start=258
  _globals['_REQUESTREPLYREQUEST']._serialized_end=338
  _globals['_REQUESTREPLYREPLY']._serialized_start=340
  _globals['_REQUESTREPLYREPLY']._serialized_end=418
  _globals['_REQUESTREPLY']._serialized_start=420
  _globals['_REQUESTREPLY']._serialized_end=546
  _globals['_LISTURISREQUEST']._serialized_start=548
  _globals['_LISTURISREQUEST']._serialized_end=565
  _globals['_LISTURISREPLY']._serialized_start=567
  _globals['_LISTURISREPLY']._serialized_end=621
  _globals['_EVENTSERVICECONFIG']._serialized_start=624
  _globals['_EVENTSERVICECONFIG']._serialized_end=976
  _globals['_EVENTSERVICECONFIG_LOGLEVEL']._serialized_start=870
  _globals['_EVENTSERVICECONFIG_LOGLEVEL']._serialized_end=976
  _globals['_EVENTSERVICECONFIGLIST']._serialized_start=978
  _globals['_EVENTSERVICECONFIGLIST']._serialized_end=1059
  _globals['_EVENTSERVICE']._serialized_start=1062
  _globals['_EVENTSERVICE']._serialized_end=1351
# @@protoc_insertion_point(module_scope)



from farm_ng.core import uri_pb2 as _uri_pb2
from farm_ng.core import event_pb2 as _event_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SubscribeRequest(_message.Message):
    __slots__ = ("uri", "every_n")
    URI_FIELD_NUMBER: _ClassVar[int]
    EVERY_N_FIELD_NUMBER: _ClassVar[int]
    uri: _uri_pb2.Uri
    every_n: int
    def __init__(self, uri: _Optional[_Union[_uri_pb2.Uri, _Mapping]] = ..., every_n: _Optional[int] = ...) -> None: ...

class SubscribeReply(_message.Message):
    __slots__ = ("event", "payload")
    EVENT_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    event: _event_pb2.Event
    payload: bytes
    def __init__(self, event: _Optional[_Union[_event_pb2.Event, _Mapping]] = ..., payload: _Optional[bytes] = ...) -> None: ...

class RequestReplyRequest(_message.Message):
    __slots__ = ("event", "payload")
    EVENT_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    event: _event_pb2.Event
    payload: bytes
    def __init__(self, event: _Optional[_Union[_event_pb2.Event, _Mapping]] = ..., payload: _Optional[bytes] = ...) -> None: ...

class RequestReplyReply(_message.Message):
    __slots__ = ("event", "payload")
    EVENT_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    event: _event_pb2.Event
    payload: bytes
    def __init__(self, event: _Optional[_Union[_event_pb2.Event, _Mapping]] = ..., payload: _Optional[bytes] = ...) -> None: ...

class RequestReply(_message.Message):
    __slots__ = ("request", "reply")
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    REPLY_FIELD_NUMBER: _ClassVar[int]
    request: RequestReplyRequest
    reply: RequestReplyReply
    def __init__(self, request: _Optional[_Union[RequestReplyRequest, _Mapping]] = ..., reply: _Optional[_Union[RequestReplyReply, _Mapping]] = ...) -> None: ...

class ListUrisRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListUrisReply(_message.Message):
    __slots__ = ("uris",)
    URIS_FIELD_NUMBER: _ClassVar[int]
    uris: _containers.RepeatedCompositeFieldContainer[_uri_pb2.Uri]
    def __init__(self, uris: _Optional[_Iterable[_Union[_uri_pb2.Uri, _Mapping]]] = ...) -> None: ...

class EventServiceConfig(_message.Message):
    __slots__ = ("name", "port", "host", "subscriptions", "uris", "args", "log_level")
    class LogLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        NOTSET: _ClassVar[EventServiceConfig.LogLevel]
        CRITICAL: _ClassVar[EventServiceConfig.LogLevel]
        FATAL: _ClassVar[EventServiceConfig.LogLevel]
        ERROR: _ClassVar[EventServiceConfig.LogLevel]
        WARNING: _ClassVar[EventServiceConfig.LogLevel]
        WARN: _ClassVar[EventServiceConfig.LogLevel]
        INFO: _ClassVar[EventServiceConfig.LogLevel]
        DEBUG: _ClassVar[EventServiceConfig.LogLevel]
    NOTSET: EventServiceConfig.LogLevel
    CRITICAL: EventServiceConfig.LogLevel
    FATAL: EventServiceConfig.LogLevel
    ERROR: EventServiceConfig.LogLevel
    WARNING: EventServiceConfig.LogLevel
    WARN: EventServiceConfig.LogLevel
    INFO: EventServiceConfig.LogLevel
    DEBUG: EventServiceConfig.LogLevel
    NAME_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    HOST_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
    URIS_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    LOG_LEVEL_FIELD_NUMBER: _ClassVar[int]
    name: str
    port: int
    host: str
    subscriptions: _containers.RepeatedCompositeFieldContainer[SubscribeRequest]
    uris: _containers.RepeatedCompositeFieldContainer[_uri_pb2.Uri]
    args: _containers.RepeatedScalarFieldContainer[str]
    log_level: EventServiceConfig.LogLevel
    def __init__(self, name: _Optional[str] = ..., port: _Optional[int] = ..., host: _Optional[str] = ..., subscriptions: _Optional[_Iterable[_Union[SubscribeRequest, _Mapping]]] = ..., uris: _Optional[_Iterable[_Union[_uri_pb2.Uri, _Mapping]]] = ..., args: _Optional[_Iterable[str]] = ..., log_level: _Optional[_Union[EventServiceConfig.LogLevel, str]] = ...) -> None: ...

class EventServiceConfigList(_message.Message):
    __slots__ = ("configs",)
    CONFIGS_FIELD_NUMBER: _ClassVar[int]
    configs: _containers.RepeatedCompositeFieldContainer[EventServiceConfig]
    def __init__(self, configs: _Optional[_Iterable[_Union[EventServiceConfig, _Mapping]]] = ...) -> None: ...



from farm_ng.core import lie_pb2 as _lie_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Pose(_message.Message):
    __slots__ = ("a_from_b", "frame_a", "frame_b", "tangent_of_b_in_a")
    A_FROM_B_FIELD_NUMBER: _ClassVar[int]
    FRAME_A_FIELD_NUMBER: _ClassVar[int]
    FRAME_B_FIELD_NUMBER: _ClassVar[int]
    TANGENT_OF_B_IN_A_FIELD_NUMBER: _ClassVar[int]
    a_from_b: _lie_pb2.Isometry3F64
    frame_a: str
    frame_b: str
    tangent_of_b_in_a: _lie_pb2.Isometry3F64Tangent
    def __init__(self, a_from_b: _Optional[_Union[_lie_pb2.Isometry3F64, _Mapping]] = ..., frame_a: _Optional[str] = ..., frame_b: _Optional[str] = ..., tangent_of_b_in_a: _Optional[_Union[_lie_pb2.Isometry3F64Tangent, _Mapping]] = ...) -> None: ...

class PoseTree(_message.Message):
    __slots__ = ("poses",)
    POSES_FIELD_NUMBER: _ClassVar[int]
    poses: _containers.RepeatedCompositeFieldContainer[Pose]
    def __init__(self, poses: _Optional[_Iterable[_Union[Pose, _Mapping]]] = ...) -> None: ...



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Uri(_message.Message):
    __slots__ = ("scheme", "authority", "path", "query")
    SCHEME_FIELD_NUMBER: _ClassVar[int]
    AUTHORITY_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    scheme: str
    authority: str
    path: str
    query: str
    def __init__(self, scheme: _Optional[str] = ..., authority: _Optional[str] = ..., path: _Optional[str] = ..., query: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/pose.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import lie_pb2 as farm__ng_dot_core_dot_lie__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x66\x61rm_ng/core/pose.proto\x12\x12\x66\x61rm_ng.core.proto\x1a\x16\x66\x61rm_ng/core/lie.proto\"\xa0\x01\n\x04Pose\x12\x32\n\x08\x61_from_b\x18\x01 \x01(\x0b\x32 .farm_ng.core.proto.Isometry3F64\x12\x0f\n\x07\x66rame_a\x18\x02 \x01(\t\x12\x0f\n\x07\x66rame_b\x18\x03 \x01(\t\x12\x42\n\x11tangent_of_b_in_a\x18\x04 \x01(\x0b\x32\'.farm_ng.core.proto.Isometry3F64Tangent\"3\n\x08PoseTree\x12\'\n\x05poses\x18\x01 \x03(\x0b\x32\x18.farm_ng.core.proto.Poseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.pose_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_POSE']._serialized_start=72
  _globals['_POSE']._serialized_end=232
  _globals['_POSETREE']._serialized_start=234
  _globals['_POSETREE']._serialized_end=285
# @@protoc_insertion_point(module_scope)



from farm_ng.core import image_pb2 as _image_pb2
from farm_ng.core import lie_pb2 as _lie_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CameraModel(_message.Message):
    __slots__ = ("image_size", "distortion_type", "params")
    IMAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    DISTORTION_TYPE_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    image_size: _image_pb2.ImageSize
    distortion_type: str
    params: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, image_size: _Optional[_Union[_image_pb2.ImageSize, _Mapping]] = ..., distortion_type: _Optional[str] = ..., params: _Optional[_Iterable[float]] = ...) -> None: ...

class CameraModels(_message.Message):
    __slots__ = ("camera_models",)
    CAMERA_MODELS_FIELD_NUMBER: _ClassVar[int]
    camera_models: _containers.RepeatedCompositeFieldContainer[CameraModel]
    def __init__(self, camera_models: _Optional[_Iterable[_Union[CameraModel, _Mapping]]] = ...) -> None: ...

class ClippingPlanes(_message.Message):
    __slots__ = ("near", "far")
    NEAR_FIELD_NUMBER: _ClassVar[int]
    FAR_FIELD_NUMBER: _ClassVar[int]
    near: float
    far: float
    def __init__(self, near: _Optional[float] = ..., far: _Optional[float] = ...) -> None: ...

class RigidCamera(_message.Message):
    __slots__ = ("intrinsics", "rig_from_camera")
    INTRINSICS_FIELD_NUMBER: _ClassVar[int]
    RIG_FROM_CAMERA_FIELD_NUMBER: _ClassVar[int]
    intrinsics: CameraModel
    rig_from_camera: _lie_pb2.Isometry3F64
    def __init__(self, intrinsics: _Optional[_Union[CameraModel, _Mapping]] = ..., rig_from_camera: _Optional[_Union[_lie_pb2.Isometry3F64, _Mapping]] = ...) -> None: ...

class GyroModel(_message.Message):
    __slots__ = ("model_type", "params")
    MODEL_TYPE_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    model_type: str
    params: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, model_type: _Optional[str] = ..., params: _Optional[_Iterable[float]] = ...) -> None: ...

class AcceleroModel(_message.Message):
    __slots__ = ("model_type", "params")
    MODEL_TYPE_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    model_type: str
    params: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, model_type: _Optional[str] = ..., params: _Optional[_Iterable[float]] = ...) -> None: ...

class ImuModel(_message.Message):
    __slots__ = ("gyro_model", "accelero_model")
    GYRO_MODEL_FIELD_NUMBER: _ClassVar[int]
    ACCELERO_MODEL_FIELD_NUMBER: _ClassVar[int]
    gyro_model: GyroModel
    accelero_model: AcceleroModel
    def __init__(self, gyro_model: _Optional[_Union[GyroModel, _Mapping]] = ..., accelero_model: _Optional[_Union[AcceleroModel, _Mapping]] = ...) -> None: ...



"""Module for the EventServiceBackend class."""
from __future__ import annotations

import asyncio
import logging

from farm_ng.core.event_client import EventClient
from farm_ng.core.event_service_pb2 import (
    EventServiceConfig,
    EventServiceConfigList,
    SubscribeRequest,
)
from farm_ng.core.uri import get_service_name
from farm_ng.core.uri_pb2 import Uri

__all__ = ["EventClientSubscriptionManager"]


class EventClientSubscriptionManager:
    """Class for the event client subscription manager.

    This class is responsible for managing the subscriptions for the service
    and maintaining the list of clients.

    Attributes:
        clients: A map of service name to EventClient.
        subscriptions: A map of service_path to SubscribeRequest.
    """

    def __init__(
        self,
        config_list: EventServiceConfigList,
        logger: logging.Logger | None = None,
    ) -> None:
        """Initializes the EventServiceBackend.

        Args:
            config_list: The configuration list for the service.
            logger: The logger for the service.
        """
        self._config_list = config_list

        # initialize the logger, ideally we would like to use the same logger as the
        # service, but we can't do that because the service is not initialized yet.
        self._logger = logger or logging.getLogger()

        # initialize a map of service name to EventClient
        self._clients: dict[str, EventClient] = self._initialize_clients(config_list)

        # initialize a map of service_path to SubscribeRequest
        self._subscriptions: dict[str, SubscribeRequest] = {}

    @property
    def clients(self) -> dict[str, EventClient]:
        """Returns a map of service name to EventClient."""
        return self._clients

    @property
    def logger(self) -> logging.Logger:
        """Returns the logger for the service."""
        return self._logger

    def _initialize_clients(
        self,
        config_list: EventServiceConfigList,
    ) -> dict[str, list[SubscribeRequest]]:
        """Initializes the clients associated with the service.

        Creates a map of service name to EventClient.

        Args:
            config_list: The configuration list for the service.

        Returns:
            A map of service name to EventClient.
        """
        # create a map of service name to EventClient
        # key: service name, value: EventClient
        clients: dict[str, EventClient] = {}

        config: EventServiceConfig
        for config in config_list.configs:
            if config.port is None:
                msg = "Port is not set for service"
                raise ValueError(msg)
            # update the service config
            clients.update({config.name: EventClient(config)})
        return clients

    def _try_register_topic(self, client_name: str, uri: Uri) -> bool:
        """Attempts to register a topic for a service.

        Args:
            client_name: The name of the client to register the topic for.
            uri: The uri to register.

        Returns:
            True if the topic was registered successfully, False otherwise.
        """
        if client_name not in self._clients:
            self.logger.warning("Service %s not found", client_name)
            return False

        # make the key for the subscription map
        service_name = get_service_name(uri)
        topic_name: str = f"{service_name}{uri.path}"

        if topic_name in self._subscriptions:
            # self.logger.warning("Topic %s already registered", topic_name)
            return False

        subscribe_request = SubscribeRequest(
            uri=Uri(
                path=uri.path,
                query=f"service_name={service_name}",
            ),
            every_n=1,
        )

        self.logger.info("Registering topic %s for client %s", topic_name, client_name)
        self._subscriptions[topic_name] = subscribe_request

        return True

    async def update_subscriptions_for_client(self, client: EventClient) -> None:
        """Updates the subscriptions for a client.

        Args:
            client: The client to update the subscriptions for.
        """
        while True:
            # get the list of uris from the client
            uris: list[Uri] = await client.list_uris()

            # register the topics
            uri: Uri
            for uri in uris:
                self._try_register_topic(client.config.name, uri)

            await asyncio.sleep(1.0)

    async def update_subscriptions(self) -> None:
        """Serves the subscriptions once to create the topics in foxglove."""
        tasks: list[asyncio.Task] = []

        client: EventClient
        for client in self._clients.values():
            tasks.append(
                asyncio.create_task(self.update_subscriptions_for_client(client)),
            )
        await asyncio.gather(*tasks)

    def get_all_uris_config_list(self, config_name: str) -> EventServiceConfigList:
        """Returns a config list with all the subscriptions for the service.

        Args:
            config_name: The name of the generated config containing all the subscriptions.

        Returns:
            A config list with all the subscriptions for the service.
        """
        # make a copy of the config list
        config_list = EventServiceConfigList()
        config_list.CopyFrom(self._config_list)

        # add the default service config and append the subscriptions
        service_config = EventServiceConfig(name=config_name)

        subscription: SubscribeRequest
        for subscription in self._subscriptions.values():
            service_config.subscriptions.append(subscription)

        config_list.configs.append(service_config)

        return config_list



from farm_ng.core import linalg_pb2 as _linalg_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RegionI32(_message.Message):
    __slots__ = ("is_empty", "min", "max")
    IS_EMPTY_FIELD_NUMBER: _ClassVar[int]
    MIN_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    is_empty: bool
    min: int
    max: int
    def __init__(self, is_empty: bool = ..., min: _Optional[int] = ..., max: _Optional[int] = ...) -> None: ...

class RegionF32(_message.Message):
    __slots__ = ("is_empty", "min", "max")
    IS_EMPTY_FIELD_NUMBER: _ClassVar[int]
    MIN_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    is_empty: bool
    min: float
    max: float
    def __init__(self, is_empty: bool = ..., min: _Optional[float] = ..., max: _Optional[float] = ...) -> None: ...

class RegionF64(_message.Message):
    __slots__ = ("is_empty", "min", "max")
    IS_EMPTY_FIELD_NUMBER: _ClassVar[int]
    MIN_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    is_empty: bool
    min: float
    max: float
    def __init__(self, is_empty: bool = ..., min: _Optional[float] = ..., max: _Optional[float] = ...) -> None: ...

class Region2F64(_message.Message):
    __slots__ = ("is_empty", "min", "max")
    IS_EMPTY_FIELD_NUMBER: _ClassVar[int]
    MIN_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    is_empty: bool
    min: _linalg_pb2.Vec2F64
    max: _linalg_pb2.Vec2F64
    def __init__(self, is_empty: bool = ..., min: _Optional[_Union[_linalg_pb2.Vec2F64, _Mapping]] = ..., max: _Optional[_Union[_linalg_pb2.Vec2F64, _Mapping]] = ...) -> None: ...

class RepeatedG0Region2F64(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: _containers.RepeatedCompositeFieldContainer[Region2F64]
    def __init__(self, value: _Optional[_Iterable[_Union[Region2F64, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/image.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x66\x61rm_ng/core/image.proto\x12\x12\x66\x61rm_ng.core.proto\"*\n\tImageSize\x12\r\n\x05width\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05\"O\n\x0bImageLayout\x12+\n\x04size\x18\x01 \x01(\x0b\x32\x1d.farm_ng.core.proto.ImageSize\x12\x13\n\x0bpitch_bytes\x18\x02 \x01(\r\"[\n\x0bPixelFormat\x12\x13\n\x0bnumber_type\x18\x01 \x01(\t\x12\x16\n\x0enum_components\x18\x02 \x01(\x05\x12\x1f\n\x17num_bytes_per_component\x18\x03 \x01(\x05\"\x80\x01\n\x08\x44ynImage\x12/\n\x06layout\x18\x01 \x01(\x0b\x32\x1f.farm_ng.core.proto.ImageLayout\x12\x35\n\x0cpixel_format\x18\x02 \x01(\x0b\x32\x1f.farm_ng.core.proto.PixelFormat\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.image_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_IMAGESIZE']._serialized_start=48
  _globals['_IMAGESIZE']._serialized_end=90
  _globals['_IMAGELAYOUT']._serialized_start=92
  _globals['_IMAGELAYOUT']._serialized_end=171
  _globals['_PIXELFORMAT']._serialized_start=173
  _globals['_PIXELFORMAT']._serialized_end=264
  _globals['_DYNIMAGE']._serialized_start=267
  _globals['_DYNIMAGE']._serialized_end=395
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import argparse
import json
import struct
import sys
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, Generator, cast

from farm_ng.core.event_pb2 import Event
from farm_ng.core.uri import uri_query_to_dict, uri_to_protobuf_type
from google.protobuf import json_format

if TYPE_CHECKING:
    from google.protobuf.message import Message

# public symbols

__all__ = [
    "build_events_dict",
    "event_has_message",
    "EventsFileReader",
    "EventLogPosition",
    "proto_from_json_file",
    "payload_to_protobuf",
]


def proto_from_json_file(
    file_path: str | Path,
    empty_proto_message: Message,
) -> Message:
    """Load a proto Message from a JSON file.

    The ``empty_proto_message`` must be the type of the proto message to load.

    Args:
        file_path (str | Path): The path to the JSON file.
        empty_proto_message (Message): The empty proto message to parse into.

    Returns:
        A proto message of the type given in empty_proto_message read from the JSON file.
    """
    if isinstance(file_path, str):
        file_path = Path(file_path).absolute()

    if not file_path.is_file():
        msg = f"Invalid file: {file_path!s}"
        raise FileNotFoundError(msg)

    with Path(file_path).open(encoding="utf-8") as file:
        json_pb = json.load(file)

    return json_format.ParseDict(json_pb, empty_proto_message)


def event_has_message(event: Event, message_type: type[Any]) -> bool:
    """Check if event has a message of type msg_type.

    Args:
        event: event_pb2.Event
        msg_type: protobuf message type

    Returns:
        bool: True if event has a message of type msg_type
    """
    if not isinstance(event, Event):
        return False
    return event.uri.query.split("&")[0].split(".")[-1] == message_type.__name__


def payload_to_protobuf(event: Event, payload: bytes) -> Message:
    """Return the protobuf message from an event and payload.

    Args:
        event: event_pb2.Event
        payload: bytes

    Returns:
        Message: the protobuf message
    """
    message_cls: type[Message] = uri_to_protobuf_type(event.uri)

    message: Message = message_cls()
    message.ParseFromString(payload)
    return message


def build_events_dict(
    events_index: list[EventLogPosition],
) -> dict[str, list[EventLogPosition]]:
    """Build a dictionary of lists of events, where the key is the path of the event.

    Args:
        events_index (list[EventLogPosition]): List of events from a log file

    Returns:
        dict[str, list[EventLogPosition]]: Dictionary of lists of events
    """
    events_dict: dict[str, list[EventLogPosition]] = defaultdict(list)
    for event_index in events_index:
        events_dict[f"/{event_index.event.uri.path}"].append(event_index)
    return events_dict


@dataclass
class EventLogPosition:
    """EventLogPosition is a dataclass that stores the event, position and reader."""

    def __init__(self, event: Event, pos: int, reader: EventsFileReader) -> None:
        """Initialize EventLogPosition.

        Args:
            event: the event described in event_pb2.Event
            pos: the position of the event in the file
            reader: the EventsFileReader instance
        """
        self.event = event
        self.pos = pos
        self.reader = reader

    def __repr__(self) -> str:
        """Return a string representation of the EventLogPosition."""
        return f"event: {self.event}\npos: {self.pos}\n"

    def read_message(self) -> Message:
        """Read the message from the event."""
        return self.reader.read_message(self)


class EventsFileReader:
    """EventsFileReader reads events from a file."""

    EVENTS_LENGTH_BYTES: int = 4

    def __init__(self, file_name: str | Path) -> None:
        """Initialize EventsFileReader.

        Args:
            file_name: path to the file
        """
        if isinstance(file_name, str):
            file_name = Path(file_name)
        self._file_name: Path = file_name.absolute()
        if not self._file_name.is_file():
            msg = f"Invalid file: {self._file_name!s}"
            raise FileNotFoundError(msg)

        self._file_stream: IO | None = None
        self._file_length: int = 0

        # store index of events in a list to seek faster
        self._events_index: list[EventLogPosition] = []

    def __enter__(self) -> EventsFileReader:
        """Open the file and return the EventsFileReader instance."""
        success: bool = self.open()
        if not success:
            msg = "Failed to open file"
            raise OSError(msg)
        return self

    def __exit__(self, type: object, value: object, traceback: object) -> None:
        """Close the file."""
        self.close()

    def __repr__(self) -> str:
        """Return a string representation of the EventsFileReader."""
        return "\n".join(
            [
                f"file_name: {self.file_name!s}",
                f"file_stream: {self._file_stream}",
                f"is_open: {self.is_open()}",
            ],
        )

    @property
    def file_length(self) -> int:
        """Return the length of the file."""
        return self._file_length

    @property
    def file_name(self) -> Path:
        """Return the path to the file."""
        return self._file_name

    @property
    def events_index(self) -> list[EventLogPosition]:
        """Return the index of events in the file."""
        return self._events_index

    def is_open(self) -> bool:
        """Return True if the file is open."""
        return not self.is_closed()

    def is_closed(self) -> bool:
        """Return True if the file is closed."""
        if self._file_stream is None:
            return True
        return self._file_stream.closed

    def open(self) -> bool:
        """Open the file and return True if successful.

        Returns:
            bool: True if successful
        """
        self._file_stream = Path(self._file_name).open("rb")
        self._file_length = self._file_name.stat().st_size
        return self.is_open()

    def close(self) -> bool:
        """Close the file and return True if successful.

        Returns:
            bool: True if successful
        """
        if self.is_closed():
            return True
        file_stream = cast(IO, self._file_stream)
        file_stream.close()
        self._file_stream = None
        return self.is_closed()

    def _reset_events_index(self) -> None:
        """Reset the events index."""
        self._events_index = []

    def read_next_event(self) -> EventLogPosition:
        """Read the next event and return it as an EventLogPosition.

        Returns:
            EventLogPosition: the event, position and reader
        """
        file_stream = cast(IO, self._file_stream)
        buffer = file_stream.read(self.EVENTS_LENGTH_BYTES)
        if len(buffer) != self.EVENTS_LENGTH_BYTES:
            msg = f"Failed to read {self.EVENTS_LENGTH_BYTES} bytes"
            raise EOFError(msg)

        # read the event length, then the event
        event_len = struct.unpack("<I", buffer)[0]
        event_bytes: bytes = file_stream.read(event_len)
        if len(event_bytes) != event_len:
            raise EOFError
        event = Event()
        event.ParseFromString(event_bytes)
        query = uri_query_to_dict(event.uri)
        event.uri.path = query.get("service_name", "") + event.uri.path
        return EventLogPosition(event, file_stream.tell(), self)

    def _skip_next_message(self, event: Event) -> None:
        """Skip the next message in the file."""
        msg_bytes: int = event.payload_length
        file_stream = cast(IO, self._file_stream)
        file_stream.seek(msg_bytes, 1)

    def _build_events_index(self) -> None:
        """Build the index of events in the file."""
        if not self.is_open():
            msg = "Reader not open. Please, use reader.open()"
            raise OSError(msg)

        file_stream = cast(IO, self._file_stream)
        file_stream.seek(0)

        # clear, if any the previous offsets
        self._reset_events_index()
        while True:
            try:
                event_log: EventLogPosition = self.read_next_event()
                self._events_index.append(event_log)
                self._skip_next_message(event_log.event)
            except EOFError:
                break

        file_stream.seek(0)

    def get_index(self) -> list[EventLogPosition]:
        """Return the index of events in the file.

        Returns:
            list[EventLogPosition]: the index of events
        """
        if len(self._events_index) == 0:
            self._build_events_index()
        return self._events_index

    def read_message(self, event_log: EventLogPosition) -> Message:
        """Read the message from the event.

        Args:
            event_log (EventLogPosition): the event log
        """
        file_stream = cast(IO, self._file_stream)
        file_stream.seek(event_log.pos, 0)
        payload: bytes = file_stream.read(event_log.event.payload_length)
        if len(payload) != event_log.event.payload_length:
            raise EOFError
        return payload_to_protobuf(event_log.event, payload)

    def read(self) -> tuple[Event, Message]:
        """Read the next event and message and return them.

        Returns:
            tuple[Event, Message]: the event and message
        """
        if not self.is_open():
            msg = "Reader not open. Please, use reader.open()"
            raise OSError(msg)

        event_log: EventLogPosition = self.read_next_event()
        return event_log.event, self.read_message(event_log)

    def read_messages(self) -> Generator[tuple[Event, Message], None, None]:
        """Read all the messages in the file and return them.

        Yields:
            Generator[tuple[Event, Message], None, None]: the event and message
        """
        if self._file_stream is None:
            msg = "Reader not open. Please, use reader.open()"
            raise OSError(msg)

        self._file_stream.seek(0)
        try:
            while True:
                event, message = self.read()
                yield event, message
        except EOFError:
            pass


def playback_command(args):
    with EventsFileReader(args.events_file) as reader:
        for event, _ in reader.read_messages():
            # Note you can find the originating service from uri query string
            query = uri_query_to_dict(event.uri)
            service_name = query.get("service_name", "")
            print(service_name + event.uri.path, event.uri.query, event.payload_length)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    sub_parsers = parser.add_subparsers()
    playback_parser = sub_parsers.add_parser("playback")
    playback_parser.add_argument("events_file")

    playback_parser.set_defaults(func=playback_command)

    args = parser.parse_args()
    if hasattr(args, "func"):

        args.func(args)
    else:
        parser.print_help()
        sys.exit(1)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/recorder.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import event_service_pb2 as farm__ng_dot_core_dot_event__service__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x61rm_ng/core/recorder.proto\x12\x12\x66\x61rm_ng.core.proto\x1a farm_ng/core/event_service.proto\"\\\n\x0fRecorderProfile\x12\x0c\n\x04name\x18\x01 \x01(\t\x12;\n\rsubscriptions\x18\x02 \x03(\x0b\x32$.farm_ng.core.proto.SubscribeRequestb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.recorder_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RECORDERPROFILE']._serialized_start=85
  _globals['_RECORDERPROFILE']._serialized_end=177
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/std.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x66\x61rm_ng/core/std.proto\x12\x12\x66\x61rm_ng.core.proto\"3\n\x0fOptionalG0Float\x12\r\n\x05value\x18\x01 \x01(\x02\x12\x11\n\thas_value\x18\x02 \x01(\x08\"4\n\x10OptionalG0Double\x12\r\n\x05value\x18\x01 \x01(\x01\x12\x11\n\thas_value\x18\x02 \x01(\x08\"3\n\x0fOptionalG0Int32\x12\r\n\x05value\x18\x01 \x01(\x05\x12\x11\n\thas_value\x18\x02 \x01(\x08\"3\n\x0fOptionalG0Int64\x12\r\n\x05value\x18\x01 \x01(\x03\x12\x11\n\thas_value\x18\x02 \x01(\x08\"2\n\x0eOptionalG0Bool\x12\r\n\x05value\x18\x01 \x01(\x08\x12\x11\n\thas_value\x18\x02 \x01(\x08\"4\n\x10OptionalG0String\x12\r\n\x05value\x18\x01 \x01(\t\x12\x11\n\thas_value\x18\x02 \x01(\x08\"%\n\x0e\x46ileSystemPath\x12\x13\n\x0bpath_string\x18\x01 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.std_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_OPTIONALG0FLOAT']._serialized_start=46
  _globals['_OPTIONALG0FLOAT']._serialized_end=97
  _globals['_OPTIONALG0DOUBLE']._serialized_start=99
  _globals['_OPTIONALG0DOUBLE']._serialized_end=151
  _globals['_OPTIONALG0INT32']._serialized_start=153
  _globals['_OPTIONALG0INT32']._serialized_end=204
  _globals['_OPTIONALG0INT64']._serialized_start=206
  _globals['_OPTIONALG0INT64']._serialized_end=257
  _globals['_OPTIONALG0BOOL']._serialized_start=259
  _globals['_OPTIONALG0BOOL']._serialized_end=309
  _globals['_OPTIONALG0STRING']._serialized_start=311
  _globals['_OPTIONALG0STRING']._serialized_end=363
  _globals['_FILESYSTEMPATH']._serialized_start=365
  _globals['_FILESYSTEMPATH']._serialized_end=402
# @@protoc_insertion_point(module_scope)



from farm_ng.core import linalg_pb2 as _linalg_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class QuaternionF64(_message.Message):
    __slots__ = ("imag", "real")
    IMAG_FIELD_NUMBER: _ClassVar[int]
    REAL_FIELD_NUMBER: _ClassVar[int]
    imag: _linalg_pb2.Vec3F64
    real: float
    def __init__(self, imag: _Optional[_Union[_linalg_pb2.Vec3F64, _Mapping]] = ..., real: _Optional[float] = ...) -> None: ...

class Rotation2F64(_message.Message):
    __slots__ = ("theta",)
    THETA_FIELD_NUMBER: _ClassVar[int]
    theta: float
    def __init__(self, theta: _Optional[float] = ...) -> None: ...

class Isometry2F64(_message.Message):
    __slots__ = ("rotation", "translation")
    ROTATION_FIELD_NUMBER: _ClassVar[int]
    TRANSLATION_FIELD_NUMBER: _ClassVar[int]
    rotation: Rotation2F64
    translation: _linalg_pb2.Vec2F64
    def __init__(self, rotation: _Optional[_Union[Rotation2F64, _Mapping]] = ..., translation: _Optional[_Union[_linalg_pb2.Vec2F64, _Mapping]] = ...) -> None: ...

class Rotation3F64(_message.Message):
    __slots__ = ("unit_quaternion",)
    UNIT_QUATERNION_FIELD_NUMBER: _ClassVar[int]
    unit_quaternion: QuaternionF64
    def __init__(self, unit_quaternion: _Optional[_Union[QuaternionF64, _Mapping]] = ...) -> None: ...

class Isometry3F64(_message.Message):
    __slots__ = ("rotation", "translation")
    ROTATION_FIELD_NUMBER: _ClassVar[int]
    TRANSLATION_FIELD_NUMBER: _ClassVar[int]
    rotation: Rotation3F64
    translation: _linalg_pb2.Vec3F64
    def __init__(self, rotation: _Optional[_Union[Rotation3F64, _Mapping]] = ..., translation: _Optional[_Union[_linalg_pb2.Vec3F64, _Mapping]] = ...) -> None: ...

class Isometry3F64Tangent(_message.Message):
    __slots__ = ("linear_velocity", "angular_velocity")
    LINEAR_VELOCITY_FIELD_NUMBER: _ClassVar[int]
    ANGULAR_VELOCITY_FIELD_NUMBER: _ClassVar[int]
    linear_velocity: _linalg_pb2.Vec3F64
    angular_velocity: _linalg_pb2.Vec3F64
    def __init__(self, linear_velocity: _Optional[_Union[_linalg_pb2.Vec3F64, _Mapping]] = ..., angular_velocity: _Optional[_Union[_linalg_pb2.Vec3F64, _Mapping]] = ...) -> None: ...



from __future__ import annotations

import struct
from pathlib import Path
from typing import IO, TYPE_CHECKING, cast

# pylint can't find Event or Uri in protobuf generated files
# https://github.com/protocolbuffers/protobuf/issues/10372
from farm_ng.core.event_pb2 import Event
from farm_ng.core.events_file_reader import payload_to_protobuf
from farm_ng.core.stamp import StampSemantics, get_monotonic_now, get_system_clock_now
from farm_ng.core.uri import make_proto_uri, uri_to_string
from google.protobuf.json_format import MessageToJson

if TYPE_CHECKING:
    from farm_ng.core.timestamp_pb2 import Timestamp
    from farm_ng.core.uri_pb2 import Uri
    from google.protobuf.message import Message

# public symbols

__all__ = [
    "EventsFileWriter",
    "proto_to_json_file",
]


def proto_to_json_file(
    file_path: str | Path,
    proto_message: Message,
    preserving_proto_field_name=False,
) -> bool:
    """Write a proto Message to a JSON file. The parent directory of the file must exist.

    Args:
        file_path (str | Path): The path to the JSON file to create / overwrite.
        proto_message (Message): The proto message to write to the JSON file.
        preserving_proto_field_name (bool): See MessageToJson in https://googleapis.dev/python/protobuf/latest/google/protobuf/json_format.html

    Returns:
        True if the file was written successfully, False otherwise.
    """
    if isinstance(file_path, str):
        file_path = Path(file_path).absolute()

    if not file_path.parent.is_dir():
        print(f"Invalid directory: {file_path.parent!s}")
        return False

    with Path(file_path).open("w", encoding="utf-8") as file:
        file.write(
            MessageToJson(
                proto_message,
                preserving_proto_field_name=preserving_proto_field_name,
            ),
        )
        file.write("\n")

    return True


class EventsFileWriter:
    """Write events to a file."""

    def __init__(
        self,
        file_base: str | Path,
        extension: str = ".bin",
        max_file_mb: int = 0,
        header_events: list[tuple[Event, bytes]] | None = None,
    ) -> None:
        """Create a new EventsFileWriter.

        Args:
            file_base: Path to and base of file name (without extension) where the events file will be logged.
            extension: Extension of the file to be logged. E.g., '.bin' or '.log'
            max_file_mb: Maximum log size in MB. Logging will roll over to new file when reached. Ignored if <= 0.
            header_events: Tuples of events & payloads to include in every split of the log file
        """
        if isinstance(file_base, str):
            file_base = Path(file_base)
        self._file_base: Path = file_base.absolute()
        self.extension: str = extension

        if not self._file_base.parent.is_dir():
            msg = f"Invalid directory: {self._file_base.parent!s}"
            raise RuntimeError(msg)

        self._file_stream: IO | None = None
        self._file_length: int = 0

        self._max_file_length = int(max(0, max_file_mb) * 1e6)
        self._file_idx: int = 0

        self._header_events: dict[str, tuple[Event, bytes]] = {}
        for (event, payload) in header_events or []:
            self.add_header_event(event, payload)

    def __enter__(self) -> EventsFileWriter:
        """Open the file for writing and return self."""
        success: bool = self.open()
        if not success:
            msg = f"Failed to open file: {self.file_name}"
            raise RuntimeError(msg)
        return self

    # pylint: disable=redefined-builtin
    def __exit__(self, type: object, value: object, traceback: object) -> None:
        """Close the file."""
        self.close()

    def __repr__(self) -> str:
        """Return a string representation of this object."""
        return (
            f"file_name: {self.file_name!s} "
            f"file_stream: {self._file_stream} "
            f"is_open: {self.is_open()} "
        )

    @property
    def max_file_length(self) -> int:
        """Max file length, in bytes."""
        return self._max_file_length

    @property
    def file_idx(self) -> int:
        """Current file number for this logging session."""
        return self._file_idx

    @property
    def file_length(self) -> int:
        """Current file length, in bytes."""
        return self._file_length

    @property
    def file_name(self) -> Path:
        """Current file name for this logging session."""
        return self._file_base.with_suffix(f".{self.file_idx:04}" + self.extension)

    def is_open(self) -> bool:
        """Return True if the file is open."""
        return not self.is_closed()

    def is_closed(self) -> bool:
        """Return True if the file is closed."""
        if self._file_stream is None:
            return True
        return self._file_stream.closed

    def _increment_file_idx(self) -> None:
        """Increment the file index."""
        self._file_idx += 1

    @property
    def header_events(self) -> dict[str, tuple[Event, bytes]]:
        """Return the dictionary of header events.

        Returns:
            dict[str, tuple[Event, bytes]]: Dictionary of header events and payloads.
                key: string representation of the uri
                value: tuple of event and payload
        """
        return self._header_events

    @property
    def header_messages(self) -> list[Message]:
        """Return the header_events, formatted as a list of protobuf messages.

        Returns:
            list[Message]: List of header messages.
        """
        return [
            payload_to_protobuf(event, payload)
            for (event, payload) in self.header_events.values()
        ]

    def add_header_event(
        self,
        event: Event,
        payload: bytes,
        write: bool = False,
    ) -> None:
        """Add a header event, and optionally writes it to the file.

        NOTE: Writing to file will fail if the file is not open.
        Args:
            event: Event to write.
            payload: Payload to write.
            write: If True, write the header event to the file. Defaults to False.
        """
        if not isinstance(event, Event):
            error_msg = f"header event must be Event, not {type(event)}"
            raise TypeError(error_msg)
        if not isinstance(payload, bytes):
            error_msg = f"header payload must be bytes, not {type(payload)}"
            raise TypeError(error_msg)
        # Once a header event is written to the file, it cannot be changed
        if uri_to_string(event.uri) not in self.header_events:
            self._header_events[uri_to_string(event.uri)] = (event, payload)
            if write:
                self.write_event_payload(event, payload)

    def write_header_events(self) -> None:
        """Write the header events to the file.

        NOTE: If the header events are too large to fit in the file,
        this will raise a RuntimeError to avoid getting stuck in an infinite loop.
        """

        true_max_file_length = self.max_file_length
        self._max_file_length = 0
        for (event, payload) in self.header_events.values():
            self.write_event_payload(event, payload)
        self._max_file_length = true_max_file_length
        if self.max_file_length and self.file_length > self.max_file_length:
            msg = f"Header events are too large to fit in a file of size {self.max_file_length}"
            raise RuntimeError(msg)

    def open(self) -> bool:
        """Open the file for writing.

        Return True if successful.
        """
        self._file_stream = Path(self.file_name).open("wb")
        self._file_length = 0
        self.write_header_events()
        return self.is_open()

    def close(self) -> bool:
        """Close the file.

        Return True if successful.
        """
        if self.is_closed():
            return True
        file_stream = cast(IO, self._file_stream)
        file_stream.close()
        self._file_stream = None
        return self.is_closed()

    def write_event_payload(self, event: Event, payload: bytes) -> None:
        if self.is_closed():
            msg = f"Event log is not open: {self.file_name}"
            raise RuntimeError(msg)

        if event.payload_length != len(payload):
            msg = f"Payload length mismatch {event.payload_length} != {len(payload)}"
            raise RuntimeError(msg)

        file_stream = cast(IO, self._file_stream)

        # note that < (little endian), I (4 bytes unsigned integer)
        event = event.SerializeToString()
        event_len: bytes = struct.pack("<I", len(event))
        self._file_length += file_stream.write(event_len)
        self._file_length += file_stream.write(event)
        self._file_length += file_stream.write(payload)

        # Rollover to new log if max file size reached
        if self.max_file_length and self.file_length > self.max_file_length:
            if not self.close():
                msg = f"Failed to close file: {self.file_name}"
                raise RuntimeError(msg)

            # Increment file index and open new file
            self._increment_file_idx()

            if not self.open():
                msg = f"Failed to open file: {self.file_name}"
                raise RuntimeError(msg)

    def _write_raw(
        self,
        uri: Uri,
        message: Message,
        timestamps: list[Timestamp],
    ) -> None:
        """Write a message to the file.

        Args:
            uri: URI of the message.
            message: Message to write.
            timestamps: List of timestamps to write.
        """

        payload = message.SerializeToString()
        event = Event(
            uri=uri,
            timestamps=timestamps,
            payload_length=len(payload),
        )
        self.write_event_payload(event, payload=payload)

    def write(
        self,
        path: str,
        message: Message,
        timestamps: list[Timestamp] | None = None,
        write_stamps: bool = True,
    ) -> None:
        """Write a message to the file.

        Args:
            path: Path to the message.
            message: Message to write.
            timestamps: List of timestamps to write.
            write_stamps: If True, append FILE_WRITE timestamps to the event when writing.
        """
        if timestamps is None:
            timestamps = []
        if write_stamps:
            timestamps.append(get_monotonic_now(semantics=StampSemantics.FILE_WRITE))
            timestamps.append(get_system_clock_now(semantics=StampSemantics.FILE_WRITE))
        uri = make_proto_uri(path=path, message=message)
        self._write_raw(uri=uri, message=message, timestamps=timestamps)



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

from farm_ng.core.timestamp_pb2 import Timestamp
from farm_ng.core.uri import get_host_name

if TYPE_CHECKING:
    from farm_ng.core.event_pb2 import Event


time_source = time


@dataclass
class StampSemantics:
    """Recommended semantics for timestamp_pb2.Timestamp timestamps.

    Example usage for when a device driver (service) receives a message from a device (e.g., a camera):
    Usage:
        stamp = get_monotonic_now(StampSemantics.DRIVER_RECEIVE)
    """

    # When a gRPC client receives a message from a gRPC server.
    CLIENT_RECEIVE = "client/receive"
    # When a gRPC client sends a message to a gRPC server.
    CLIENT_SEND = "client/send"
    # When a device (e.g., camera) takes a measurement.
    DEVICE_SAMPLE = "device/sample"
    # When a device driver (gRPC service) receives a measurement/message from a device.
    DRIVER_RECEIVE = "driver/receive"
    # When a device driver (gRPC service) sends a message to a device.
    DRIVER_TRANSMIT = "driver/transmit"
    # When a message is written to a file on disk.
    FILE_WRITE = "file/write"
    # When a gRPC service (which may be a device driver) receives a message
    # from a gRPC client or another gRPC service. This is distinct from DRIVER_RECEIVE,
    # which is when a device driver receives a message from a device.
    SERVICE_RECEIVE = "service/receive"
    # When a gRPC service (which may be a device driver) sends a message
    SERVICE_SEND = "service/send"


def get_monotonic_now(semantics: str) -> Timestamp:
    return Timestamp(
        stamp=time_source.monotonic(),
        clock_name=get_host_name() + "/monotonic",
        semantics=semantics,
    )


def get_system_clock_now(semantics: str) -> Timestamp:
    return Timestamp(
        stamp=time_source.time(),
        clock_name=get_host_name() + f"/system_clock/{time_source.localtime().tm_zone}",
        semantics=semantics,
    )


def timestamp_from_monotonic(semantics: str, stamp: float) -> Timestamp:
    return Timestamp(
        stamp=stamp,
        clock_name=get_host_name() + "/monotonic",
        semantics=semantics,
    )


def get_stamp_by_semantics_and_clock_type(
    event: Event,
    semantics: str,
    clock_type: str,
) -> float | None:
    """Utility function to get the stamp of a Timestamp by semantics.

    Args:
        event: farm-ng protobuf event
        semantics: semantics of the timestamp to get (see StampSemantics)
        clock_type: clock type of the timestamp to get (e.g., monotonic, system_clock)
    """
    for timestamp in event.timestamps:
        if timestamp.semantics == semantics and clock_type in timestamp.clock_name:
            return timestamp.stamp
    return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/geometry.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import linalg_pb2 as farm__ng_dot_core_dot_linalg__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x61rm_ng/core/geometry.proto\x12\x12\x66\x61rm_ng.core.proto\x1a\x19\x66\x61rm_ng/core/linalg.proto\"8\n\x0bUnitVec3F64\x12)\n\x04vec3\x18\x01 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64\"Q\n\x0eHyperplane3F64\x12/\n\x06normal\x18\x01 \x01(\x0b\x32\x1f.farm_ng.core.proto.UnitVec3F64\x12\x0e\n\x06offset\x18\x02 \x01(\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.geometry_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_UNITVEC3F64']._serialized_start=78
  _globals['_UNITVEC3F64']._serialized_end=134
  _globals['_HYPERPLANE3F64']._serialized_start=136
  _globals['_HYPERPLANE3F64']._serialized_end=217
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/lie.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import linalg_pb2 as farm__ng_dot_core_dot_linalg__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x66\x61rm_ng/core/lie.proto\x12\x12\x66\x61rm_ng.core.proto\x1a\x19\x66\x61rm_ng/core/linalg.proto\"H\n\rQuaternionF64\x12)\n\x04imag\x18\x01 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64\x12\x0c\n\x04real\x18\x02 \x01(\x01\"\x1d\n\x0cRotation2F64\x12\r\n\x05theta\x18\x01 \x01(\x01\"t\n\x0cIsometry2F64\x12\x32\n\x08rotation\x18\x01 \x01(\x0b\x32 .farm_ng.core.proto.Rotation2F64\x12\x30\n\x0btranslation\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec2F64\"J\n\x0cRotation3F64\x12:\n\x0funit_quaternion\x18\x01 \x01(\x0b\x32!.farm_ng.core.proto.QuaternionF64\"t\n\x0cIsometry3F64\x12\x32\n\x08rotation\x18\x01 \x01(\x0b\x32 .farm_ng.core.proto.Rotation3F64\x12\x30\n\x0btranslation\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64\"\x82\x01\n\x13Isometry3F64Tangent\x12\x34\n\x0flinear_velocity\x18\x01 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64\x12\x35\n\x10\x61ngular_velocity\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.lie_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_QUATERNIONF64']._serialized_start=73
  _globals['_QUATERNIONF64']._serialized_end=145
  _globals['_ROTATION2F64']._serialized_start=147
  _globals['_ROTATION2F64']._serialized_end=176
  _globals['_ISOMETRY2F64']._serialized_start=178
  _globals['_ISOMETRY2F64']._serialized_end=294
  _globals['_ROTATION3F64']._serialized_start=296
  _globals['_ROTATION3F64']._serialized_end=370
  _globals['_ISOMETRY3F64']._serialized_start=372
  _globals['_ISOMETRY3F64']._serialized_end=488
  _globals['_ISOMETRY3F64TANGENT']._serialized_start=491
  _globals['_ISOMETRY3F64TANGENT']._serialized_end=621
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/sensor.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import image_pb2 as farm__ng_dot_core_dot_image__pb2
from farm_ng.core import lie_pb2 as farm__ng_dot_core_dot_lie__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x66\x61rm_ng/core/sensor.proto\x12\x12\x66\x61rm_ng.core.proto\x1a\x18\x66\x61rm_ng/core/image.proto\x1a\x16\x66\x61rm_ng/core/lie.proto\"i\n\x0b\x43\x61meraModel\x12\x31\n\nimage_size\x18\x01 \x01(\x0b\x32\x1d.farm_ng.core.proto.ImageSize\x12\x17\n\x0f\x64istortion_type\x18\x02 \x01(\t\x12\x0e\n\x06params\x18\x03 \x03(\x01\"F\n\x0c\x43\x61meraModels\x12\x36\n\rcamera_models\x18\x01 \x03(\x0b\x32\x1f.farm_ng.core.proto.CameraModel\"+\n\x0e\x43lippingPlanes\x12\x0c\n\x04near\x18\x01 \x01(\x02\x12\x0b\n\x03\x66\x61r\x18\x02 \x01(\x02\"}\n\x0bRigidCamera\x12\x33\n\nintrinsics\x18\x01 \x01(\x0b\x32\x1f.farm_ng.core.proto.CameraModel\x12\x39\n\x0frig_from_camera\x18\x02 \x01(\x0b\x32 .farm_ng.core.proto.Isometry3F64\"/\n\tGyroModel\x12\x12\n\nmodel_type\x18\x01 \x01(\t\x12\x0e\n\x06params\x18\x02 \x03(\x01\"3\n\rAcceleroModel\x12\x12\n\nmodel_type\x18\x01 \x01(\t\x12\x0e\n\x06params\x18\x02 \x03(\x01\"x\n\x08ImuModel\x12\x31\n\ngyro_model\x18\x01 \x01(\x0b\x32\x1d.farm_ng.core.proto.GyroModel\x12\x39\n\x0e\x61\x63\x63\x65lero_model\x18\x02 \x01(\x0b\x32!.farm_ng.core.proto.AcceleroModelb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.sensor_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CAMERAMODEL']._serialized_start=99
  _globals['_CAMERAMODEL']._serialized_end=204
  _globals['_CAMERAMODELS']._serialized_start=206
  _globals['_CAMERAMODELS']._serialized_end=276
  _globals['_CLIPPINGPLANES']._serialized_start=278
  _globals['_CLIPPINGPLANES']._serialized_end=321
  _globals['_RIGIDCAMERA']._serialized_start=323
  _globals['_RIGIDCAMERA']._serialized_end=448
  _globals['_GYROMODEL']._serialized_start=450
  _globals['_GYROMODEL']._serialized_end=497
  _globals['_ACCELEROMODEL']._serialized_start=499
  _globals['_ACCELEROMODEL']._serialized_end=550
  _globals['_IMUMODEL']._serialized_start=552
  _globals['_IMUMODEL']._serialized_end=672
# @@protoc_insertion_point(module_scope)



"""EventServiceRecorder and RecorderService.

Run the event_service which starts a simple test publisher:
$ python -m farm_ng.core.event_service

Run the event_recorder which subscribes to the test publisher and logs the events to a file:
$ python -m farm_ng.core.event_service_recorder record --service-config=config.json --config-name=record_all foo

NOTE: The config file has an EventServiceConfig with name "record_all" which subscribes to the test publisher

Play back the log file:
$ python -m farm_ng.core.events_file_reader playback foo.0000.bin

Try running the event_service_recorder in service mode.

Start the service:
$ python -m farm_ng.core.event_service_recorder service --service-config config.json --service-name recorder

In a separate terminal, start recording:
$ python -m farm_ng.core.event_service_recorder start --service-config config.json  --service-name recorder

Then stop recording.
$ python -m farm_ng.core.event_service_recorder stop --service-config config.json  --service-name recorder
"""
from __future__ import annotations

import argparse
import asyncio
import logging
import sys
from collections import deque
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING

import grpc
from farm_ng.core import event_pb2
from farm_ng.core.event_client import EventClient
from farm_ng.core.event_service import (
    EventServiceGrpc,
    add_service_parser,
    load_service_config,
)
from farm_ng.core.event_service_pb2 import (
    EventServiceConfig,
    EventServiceConfigList,
    RequestReplyRequest,
    SubscribeRequest,
)
from farm_ng.core.events_file_reader import payload_to_protobuf, proto_from_json_file
from farm_ng.core.events_file_writer import EventsFileWriter
from farm_ng.core.stamp import StampSemantics, get_monotonic_now
from farm_ng.core.uri import get_host_name, uri_query_to_dict, uri_to_string
from google.protobuf.empty_pb2 import Empty
from google.protobuf.wrappers_pb2 import StringValue

if TYPE_CHECKING:
    from google.protobuf.message import Message

__all__ = ["EventServiceRecorder", "RecorderService"]


class EventServiceRecorder:
    """Service that subscribes to a list of services and records the events to a file."""

    # the maximum size of the queue
    QUEUE_MAX_SIZE: int = 50

    def __init__(
        self,
        service_name: str,
        config_list: EventServiceConfigList,
        header_events: list[tuple[event_pb2.Event, bytes]] | None = None,
    ) -> None:
        """Initializes the service.

        Args:
            service_name (str): the name of the service.
            config_list (EventServiceConfigList): the configuration data structure.
            header_events (list[tuple[event_pb2.Event, bytes]], optional): the initial list header events.
        """
        self.service_name: str = service_name
        self.config_list: EventServiceConfigList = config_list
        self.recorder_config: EventServiceConfig | None = None

        # the service logger
        self._logger: logging.Logger = logging.getLogger(service_name)

        # map of service name to EventClient
        self._clients: dict[str, EventClient] = {}

        # find the recorder config and create the clients
        # Here we reuse a config for the subscriptions, from the recorder_config
        # and create the clients for the other services (with port names)
        config: EventServiceConfig
        for config in self.config_list.configs:
            if self.service_name == config.name:
                self.recorder_config = config
            elif config.port != 0:
                self._clients[config.name] = EventClient(config)

        if self.recorder_config is None:
            msg = f"service {self.service_name} not found in config list"
            raise ValueError(msg)

        # the queue to store the events
        self.record_queue: asyncio.Queue[tuple[event_pb2.Event, bytes]] = asyncio.Queue(
            maxsize=self.QUEUE_MAX_SIZE,
        )
        self.header_deque: deque[tuple[event_pb2.Event, bytes]] = deque()
        if header_events is not None:
            for event, payload in header_events:
                self.add_header_event(event, payload)

    def add_header_event(self, event: event_pb2.Event, payload: bytes) -> None:
        """Add a header event to the header_deque.

        Args:
            event: Event to add.
            payload: Payload to add.
        """
        if not isinstance(event, event_pb2.Event):
            error_msg = f"header event must be Event, not {type(event)}"
            raise TypeError(error_msg)
        if not isinstance(payload, bytes):
            error_msg = f"header payload must be bytes, not {type(payload)}"
            raise TypeError(error_msg)
        self.header_deque.append((event, payload))

    @property
    def logger(self) -> logging.Logger:
        """Returns the service logger."""
        return self._logger

    @property
    def clients(self) -> dict[str, EventClient]:
        """Returns the map of service name to EventClient."""
        return self._clients

    async def record(
        self,
        file_base: str | Path,
        extension: str = ".bin",
        max_file_mb: int = 0,
    ) -> None:
        """Records the events that arrive in the queue to a file.

        Args:
            file_base (str | Path): the base name of the file.
            extension (str, optional): the extension of the file. Defaults to ".bin".
            max_file_mb (int, optional): the maximum size of the file in MB. Defaults to 0.
        """
        with EventsFileWriter(
            file_base,
            extension,
            max_file_mb,
            header_events=list(self.header_deque),
        ) as writer:
            self.header_deque.clear()
            event: event_pb2.Event
            payload: bytes
            while True:
                # Add any header events added during recording
                while self.header_deque:
                    event, payload = self.header_deque.popleft()
                    writer.add_header_event(event, payload, write=True)
                # await a new event and payload, and write it to the file
                event, payload = await self.record_queue.get()
                event.timestamps.append(
                    get_monotonic_now(semantics=StampSemantics.FILE_WRITE),
                )
                writer.write_event_payload(event, payload)

    async def subscribe(
        self,
        client: EventClient,
        subscription: SubscribeRequest,
    ) -> None:
        """Subscribes to a service and puts the events in the queue.

        Args:
            client (EventClient): the client to subscribe to.
            subscription (SubscribeRequest): the subscription request.
        """
        event: event_pb2.Event
        payload: bytes
        async for event, payload in client.subscribe(subscription, decode=False):
            if event.uri.path.startswith("/header"):
                # Handle header events - typically these will be metadata or calibrations
                self.add_header_event(event, payload)
            try:
                self.record_queue.put_nowait((event, payload))
            except asyncio.QueueFull:
                self.logger.warning("Queue full, dropping event %s", event)
                continue

    async def subscribe_and_record(
        self,
        file_base: str | Path,
        extension: str = ".bin",
        max_file_mb: int = 0,
    ) -> None:
        """Subscribes to a list of services and records the events to a file.

        Args:
            file_base (str | Path): the base name of the file.
            extension (str, optional): the extension of the file. Defaults to ".bin".
            max_file_mb (int, optional): the maximum size of the file in MB. Defaults to 0.
        """
        if self.recorder_config is None:
            msg = "recorder_config is None"
            raise ValueError(msg)

        # create the tasks list. the first task is the record task
        async_tasks: list[asyncio.Task] = [
            asyncio.create_task(
                self.record(
                    file_base=file_base,
                    extension=extension,
                    max_file_mb=max_file_mb,
                ),
            ),
        ]

        # add the subscribe tasks based on the subscriptions in the recorder config
        subscription: SubscribeRequest
        for subscription in self.recorder_config.subscriptions:
            query: dict[str, str] = uri_query_to_dict(uri=subscription.uri)
            query_service_name: str = query["service_name"]
            if query_service_name not in self.clients:
                self.logger.warning("Invalid subscription: %s", query_service_name)
                continue
            client: EventClient = self.clients[query_service_name]
            async_tasks.append(
                asyncio.create_task(self.subscribe(client, subscription)),
            )
        try:
            await asyncio.gather(*async_tasks)
        except asyncio.CancelledError:
            self.logger.debug("cancelled recording")
        finally:
            self.logger.info("done recording")


# the format of the file name
# Example: 2021_08_31_15_54_00_000000
DATETIME_FORMAT: str = "%Y_%m_%d_%H_%M_%S_%f"

# the default timezone
DEFAULT_TIMEZONE = timezone.utc


def get_file_name_base() -> str:
    """Returns the base name of the file.

    The base name is the current date and time, and the host name.

    Returns:
        str: the base name of the file.

    Example:
        >>> get_file_name_base()
        '2021_08_31_15_54_00_000000_ubuntu'
    """
    return (
        datetime.now(DEFAULT_TIMEZONE).strftime(DATETIME_FORMAT) + "_" + get_host_name()
    )


class RecorderService:
    """Service that subscribes to a list of services and records the events to a file.

    - This service will record events to a file when it receives a request to start.
    - It will stop recording when it receives a request to stop.
    - It will publish the progress of recording.
    - It will only record one recording at a time.
    """

    def __init__(self, event_service: EventServiceGrpc) -> None:
        """Initializes the service.

        Args:
            event_service (EventServiceGrpc): the event service.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument("--data-dir", required=True)
        args = parser.parse_args(event_service.config.args)

        self._data_dir: Path = Path(args.data_dir)
        self._data_dir.mkdir(parents=True, exist_ok=True)
        if not self._data_dir.exists():
            msg = f"{self._data_dir} does not exist"
            raise OSError(msg)

        # set the request reply handler
        self._event_service = event_service
        self._event_service.request_reply_handler = self._request_reply_handler

        # the recorder event service
        self._recorder: EventServiceRecorder | None = None

        # the recorder task
        self._recorder_task: asyncio.Task | None = None

        # For tracking header events (e.g. metadata, calibrations) to be logged in the recordings
        self.header_events: dict[str, tuple[event_pb2.Event, bytes]] = {}

    # public methods

    async def start_recording(
        self,
        file_base: Path,
        config_list: EventServiceConfigList,
        config_name: str | None = None,
    ) -> None:
        """Starts recording the events to a file.

        Args:
            file_base (Path): the base name of the file.
            config_list (EventServiceConfigList): the configuration data structure.
            config_name (str, optional): the name of the config. Defaults to None,
                                            in which case "record_default" is used.
        """
        # stop recording if it is already running
        if self._recorder is not None:
            await self.stop_recording()

        # start recording
        self._recorder = EventServiceRecorder(
            config_name or "record_default",
            config_list,
            list(self.header_events.values()),
        )
        if self._recorder.recorder_config is None:
            msg = "recorder_config is None"
            raise ValueError(msg)

        # Handle if the max_file_mb is set in the record_config args
        # Expectation: single string of "--max-file-mb=500" or "--max-file-mb 500"
        max_file_mb: int = 0
        for arg in self._recorder.recorder_config.args:
            if arg.startswith("--max-file-mb"):
                try:
                    eq_arg = arg.replace(" ", "=").strip()
                    max_file_mb = int(eq_arg.split("=")[1])
                    self._recorder.logger.info("Setting max_file_mb to %s", max_file_mb)
                except ValueError:
                    self._recorder.logger.exception(
                        "Failed to parse max_file_mb from %s",
                        eq_arg,
                    )

        self._recorder_task = asyncio.create_task(
            self._recorder.subscribe_and_record(
                file_base=file_base,
                max_file_mb=max_file_mb,
            ),
        )

        def _safe_done_callback(
            future: asyncio.Task,
            recorder_task: asyncio.Task | None,
        ) -> None:
            try:
                future.result()
            except asyncio.CancelledError:
                self._event_service.logger.debug("cancelled recording task")
            finally:
                del future
                del recorder_task
                self._recorder = None

        self._recorder_task.add_done_callback(
            lambda f: _safe_done_callback(f, self._recorder_task),
        )

    async def stop_recording(self) -> None:
        """Stops recording the events to a file."""
        # do nothing if not recording
        if self._recorder is None or self._recorder_task is None:
            return

        self._recorder.logger.info("stopping recording")

        # cancel the task and set the recorder to None
        self._recorder_task.cancel()

    # private methods

    async def _request_reply_handler(
        self,
        request: RequestReplyRequest,
    ) -> Message:
        """Handles the request reply.

        Args:
            event_service (EventServiceGrpc): the event service.
            request (RequestReplyRequest): the request.

        Returns:
            Message: the response message.
        """
        cmd: str = request.event.uri.path
        # Ensure path was sent with a leading slash
        if not cmd.startswith("/"):
            cmd = "/" + cmd

        if cmd.startswith("/start"):
            # config_name is the optional part of the path after "/start/"
            config_name: str | None = cmd[7:] or None

            config_list: EventServiceConfigList = payload_to_protobuf(
                request.event,
                request.payload,
            )
            self._event_service.logger.info("/start %s: %s", config_name, config_list)
            file_base = self._data_dir.joinpath(get_file_name_base())
            await self.start_recording(file_base, config_list, config_name)
            return StringValue(value=str(file_base))
        if cmd == "/stop":
            self._event_service.logger.info("/stop")
            await self.stop_recording()
        elif cmd.startswith("/header"):
            header_detail: str = cmd[8:] or ""
            if not header_detail:
                msg = "/header command requires a specifier, e.g. '/header/metadata'"
                raise ValueError(msg)
            self._event_service.logger.info(
                "header:\n%s",
                payload_to_protobuf(request.event, request.payload),
            )
            self._event_service.logger.info("with uri:\n%s", request.event.uri)
            self.add_header_event(request.event, request.payload)
        elif cmd == "/clear_headers":
            self._event_service.logger.info("/clear_headers")
            self.header_events.clear()
            if self._recorder is not None:
                self._recorder.header_deque.clear()
        return Empty()

    def add_header_event(self, event: event_pb2.Event, payload: bytes) -> None:
        """Adds a header event to the header_events list.

        If this is a duplicate (per URI), it will replace the existing message.

        Args:
            event (event_pb2.Event): the event.
            payload (bytes): the payload.
        """
        if not isinstance(event, event_pb2.Event):
            error_msg = f"header event must be Event, not {type(event)}"
            raise TypeError(error_msg)
        if not isinstance(payload, bytes):
            error_msg = f"header payload must be bytes, not {type(payload)}"
            raise TypeError(error_msg)
        self.header_events[uri_to_string(event.uri)] = (event, payload)
        if self._recorder is not None:
            self._recorder.add_header_event(event, payload)


def service_command(_args):
    config_list, service_config = load_service_config(args)
    event_service: EventServiceGrpc = EventServiceGrpc(
        grpc.aio.server(),
        service_config,
    )
    RecorderService(event_service)

    async def job():
        # define the async tasks
        async_tasks: list[asyncio.Task] = []

        # add the service task
        async_tasks.append(event_service.serve())

        # run the tasks
        await asyncio.gather(*async_tasks)

    asyncio.get_event_loop().run_until_complete(job())


def client_start_command(_args):
    config_list, service_config = load_service_config(args)
    loop = asyncio.get_event_loop()
    config_name: str = args.config_name

    async def job():
        reply = await EventClient(service_config).request_reply(
            f"/start/{config_name}",
            config_list,
        )
        print(payload_to_protobuf(reply.event, reply.payload))

    loop.run_until_complete(job())


def client_stop_command(_args):
    config_list, service_config = load_service_config(args)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(EventClient(service_config).request_reply("/stop", Empty()))


def record_command(_args):
    async def job():
        configs = proto_from_json_file(_args.service_config, EventServiceConfigList())
        recorder = EventServiceRecorder(_args.config_name, config_list=configs)
        await recorder.subscribe_and_record(
            _args.output,
            _args.extension,
            _args.max_file_mb,
        )

    asyncio.get_event_loop().run_until_complete(job())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    sub_parsers = parser.add_subparsers()
    service_parser = sub_parsers.add_parser("service")
    add_service_parser(service_parser)
    service_parser.add_argument("--out-dir", default="/tmp/")  # noqa: S108
    service_parser.set_defaults(func=service_command)

    client_parser = sub_parsers.add_parser("start")
    add_service_parser(client_parser)
    client_parser.add_argument("--config-name", default="record_default")
    client_parser.set_defaults(func=client_start_command)

    client_parser = sub_parsers.add_parser("stop")
    add_service_parser(client_parser)
    client_parser.set_defaults(func=client_stop_command)

    record_parser = sub_parsers.add_parser("record")
    record_parser.add_argument("--service-config", required=True, type=str)
    record_parser.add_argument("--output")
    record_parser.add_argument("--extension", default=".bin")
    record_parser.add_argument("--max_file_mb", default=0, type=int)
    record_parser.add_argument("--config-name", default="record_default")

    record_parser.set_defaults(func=record_command)

    args = parser.parse_args()
    if hasattr(args, "func"):

        args.func(args)
    else:
        parser.print_help()
        sys.exit(1)



from __future__ import annotations

import importlib
import platform
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from farm_ng.core import uri_pb2

if TYPE_CHECKING:
    from google.protobuf.message import Message


__all__ = [
    "PlatformConfig",
    "get_host_name",
    "get_authority",
    "make_proto_uri",
    "uri_to_string",
    "string_to_uri",
    "uri_query_to_dict",
    "uri_to_protobuf_type",
]


def _get_platform_node() -> str:
    """Return the platform node name.

    Example:
        >>> get_platform_node()
        'farm-ng-1'
    """
    return platform.node()


@dataclass
class PlatformConfig:
    """Platform configuration."""

    host_name: str = field(default_factory=_get_platform_node)


# https://vald-phoenix.github.io/pylint-errors/plerr/errors/variables/W0603.html
platform_config = PlatformConfig()


def set_host_name(name: str) -> None:
    """Set the host name."""
    # global HOST_NAME
    platform_config.host_name = name


def get_host_name() -> str:
    """Return the host name.

    Example:
        >>> get_host_name()
        'farm-ng-1'
    """
    return platform_config.host_name


def get_authority() -> str:
    """Return the authority.

    Example:
        >>> get_authority()
        'farm-ng-1'
    """
    return get_host_name()


def make_proto_uri(
    path: str,
    message: Message,
    service_name: str | None = None,
) -> uri_pb2.Uri:
    """Return a protobuf uri from protobuf message.

    Args:
        path: the semantic path.
        message: the protobuf message.

    Returns:
        uri_pb2.Uri: the protobuf uri.

    Example:
        >>> from farm_ng.core import timestamp_pb2
        >>> stamp = timestamp_pb2.Timestamp(
        ...     stamp=1.2,
        ...     clock_name="clock0",
        ...     semantics="test/proto",
        ... )
        >>> make_proto_uri("tik/tok", stamp)
        scheme: "protobuf"
        authority: "farm-ng-1"
        path: "tik/tok"
        query: "type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto"
    """
    service_query = ""
    if service_name is not None:
        service_query = f"&service_name={service_name}"
    return uri_pb2.Uri(
        scheme="protobuf",
        authority=get_authority(),
        path=path,
        query=f"type={message.DESCRIPTOR.full_name}&pb={message.DESCRIPTOR.file.name}{service_query}",
    )


def get_service_name(uri: uri_pb2.Uri) -> str | None:
    """Return the service name from a protobuf uri.

    Args:
        uri: uri_pb2.Uri

    Returns:
        str | None: the service name.

    Example:
        >>> uri = Uri()
        >>> uri.query = "type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto&service_name=service"
        >>> get_service_name(uri)
        'service'
    """
    return uri_query_to_dict(uri).get("service_name", None)


def uri_to_string(uri: uri_pb2.Uri) -> str:
    """Return a string from a protobuf uri.

    Args:
        uri: uri_pb2.Uri

    Returns:
        str: the string representation of the uri.

    Example:
        >>> uri = Uri()
        >>> uri.scheme = "protobuf"
        >>> uri.authority = "farm-ng-1"
        >>> uri.query = "type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto"
        >>> uri.path = "/stamp_stream"
        >>> uri_to_string(uri)
        'protobuf://farm-ng-1//stamp_stream?type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto'
    """
    return f"{uri.scheme}://{uri.authority}/{uri.path}?{uri.query}"


def string_to_uri(string: str) -> uri_pb2.Uri:
    """Return a protobuf uri from a string.

    Args:
        string: the string representation of the uri.

    Returns:
        uri_pb2.Uri: the protobuf uri.

    Example:
        >>> string = "protobuf://farm-ng-1//?type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto"
        >>> string_to_uri(string)
        scheme: "protobuf"
        authority: "farm-ng-1"
        path: ""
        query: "type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto"
    """
    scheme_name, remainder = string.split("://")
    remainder, query = remainder.split("?")
    authority_path = remainder.split("/")
    authority = authority_path[0]
    path = "/".join(authority_path[1:])
    return uri_pb2.Uri(scheme=scheme_name, authority=authority, path=path, query=query)


def uri_query_to_dict(uri: uri_pb2.Uri) -> dict[str, str]:
    """Return a dictionary from a uri query.

    Args:
        uri: uri_pb2.Uri

    Returns:
        dict[str, str]: the dictionary representation of the uri query.

    Example:
        >>> uri = Uri()
        >>> uri.query = "type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto"
        >>> uri_query_to_dict(uri)
        {'type': 'farm_ng.core.proto.Timestamp', 'pb': 'farm_ng/core/timestamp.proto'}
    """
    return dict([x.split("=") for x in uri.query.split("&")])


def uri_to_protobuf_type(uri: uri_pb2.Uri) -> type[Message]:
    """Return the protobuf type from an event.

    Args:
        event: event_pb2.Event

    Returns:
        Type[Message]: the protobuf type

    Example:
        >>> event = Event()
        >>> event.uri.query = "type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto"
        >>> event_to_protobuf_type(event)
        <class 'farm_ng.core.timestamp_pb2.Timestamp'>
    """
    name: str
    package: str
    name, package = _parse_protobuf_descriptor(uri)
    return getattr(importlib.import_module(package), name)


# private functions

# parse a proto descriptor and extract the message name and package.
# See: https://developers.google.com/protocol-buffers/docs/reference/python-generated#invocation
# NOTE: the descriptor comes in the shape of:
# `type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto`


def _parse_protobuf_descriptor(uri: uri_pb2.Uri) -> tuple[str, str]:
    """Parse a protobuf descriptor and extract the message name and package.

    Args:
        uri: uri_pb2.Uri

    Returns:
        tuple[str, str]: message name and package

    Example:
        >>> uri = Uri()
        >>> uri.query = "type=farm_ng.core.proto.Timestamp&pb=farm_ng/core/timestamp.proto"
        >>> parse_protobuf_descriptor(uri)
        ('Timestamp', 'farm_ng.core.timestamp_pb2')
    """
    if uri.scheme != "protobuf":
        msg = f"Invalid uri scheme: {uri.scheme}"
        raise ValueError(msg)

    # convert the query string to a dictionary and extract the type and protobuf file name
    query: dict[str, str] = uri_query_to_dict(uri)
    type_split: str | None = query.get("type", None)
    pb_split: str | None = query.get("pb", None)

    if type_split is None:
        msg = f"Invalid uri query: {uri.query}"
        raise ValueError(msg)

    if pb_split is None:
        msg = f"Invalid uri query: {uri.query}"
        raise ValueError(msg)

    # the message name is the last part of the type
    type_name: str = type_split.split(".")[-1]

    # parse the pb file location and convert to python module extension
    package_name: str = pb_split.replace(".proto", "_pb2").replace("/", ".")

    return type_name, package_name



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/calculus.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import linalg_pb2 as farm__ng_dot_core_dot_linalg__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x61rm_ng/core/calculus.proto\x12\x12\x66\x61rm_ng.core.proto\x1a\x19\x66\x61rm_ng/core/linalg.proto\"7\n\tRegionI32\x12\x10\n\x08is_empty\x18\x01 \x01(\x08\x12\x0b\n\x03min\x18\x02 \x01(\x05\x12\x0b\n\x03max\x18\x03 \x01(\x05\"7\n\tRegionF32\x12\x10\n\x08is_empty\x18\x01 \x01(\x08\x12\x0b\n\x03min\x18\x02 \x01(\x02\x12\x0b\n\x03max\x18\x03 \x01(\x02\"7\n\tRegionF64\x12\x10\n\x08is_empty\x18\x01 \x01(\x08\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\"r\n\nRegion2F64\x12\x10\n\x08is_empty\x18\x01 \x01(\x08\x12(\n\x03min\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec2F64\x12(\n\x03max\x18\x03 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec2F64\"E\n\x14RepeatedG0Region2F64\x12-\n\x05value\x18\x01 \x03(\x0b\x32\x1e.farm_ng.core.proto.Region2F64b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.calculus_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_REGIONI32']._serialized_start=78
  _globals['_REGIONI32']._serialized_end=133
  _globals['_REGIONF32']._serialized_start=135
  _globals['_REGIONF32']._serialized_end=190
  _globals['_REGIONF64']._serialized_start=192
  _globals['_REGIONF64']._serialized_end=247
  _globals['_REGION2F64']._serialized_start=249
  _globals['_REGION2F64']._serialized_end=363
  _globals['_REPEATEDG0REGION2F64']._serialized_start=365
  _globals['_REPEATEDG0REGION2F64']._serialized_end=434
# @@protoc_insertion_point(module_scope)



from farm_ng.core import linalg_pb2 as _linalg_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class UnitVec3F64(_message.Message):
    __slots__ = ("vec3",)
    VEC3_FIELD_NUMBER: _ClassVar[int]
    vec3: _linalg_pb2.Vec3F64
    def __init__(self, vec3: _Optional[_Union[_linalg_pb2.Vec3F64, _Mapping]] = ...) -> None: ...

class Hyperplane3F64(_message.Message):
    __slots__ = ("normal", "offset")
    NORMAL_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    normal: UnitVec3F64
    offset: float
    def __init__(self, normal: _Optional[_Union[UnitVec3F64, _Mapping]] = ..., offset: _Optional[float] = ...) -> None: ...



"""
Example Usage:

Connect to the test_service and print out the messages it receives:
$ python -m farm_ng.core.event_client --service-config config.json --service-name test_service
"""
from __future__ import annotations

import argparse
import asyncio
import logging
from typing import TYPE_CHECKING, AsyncIterator, Protocol

import grpc
from farm_ng.core import event_service_pb2_grpc
from farm_ng.core.event_pb2 import Event
from farm_ng.core.event_service import add_service_parser, load_service_config
from farm_ng.core.event_service_pb2 import (
    EventServiceConfig,
    ListUrisReply,
    ListUrisRequest,
    RequestReplyReply,
    RequestReplyRequest,
    SubscribeReply,
    SubscribeRequest,
)
from farm_ng.core.events_file_reader import payload_to_protobuf
from farm_ng.core.events_file_writer import make_proto_uri
from farm_ng.core.stamp import (
    StampSemantics,
    get_monotonic_now,
    get_stamp_by_semantics_and_clock_type,
    get_system_clock_now,
)

if TYPE_CHECKING:
    from farm_ng.core.timestamp_pb2 import Timestamp
    from farm_ng.core.uri_pb2 import Uri
    from google.protobuf.message import Message

from .event_service_metrics import EventServiceHealthMetrics

logging.basicConfig(level=logging.INFO)


__all__ = ["EventClient"]


class EventClientProtocol(Protocol):
    """Protocol for the gRPC streaming object.

    This is used to make the code more testable.
    """

    async def read(self) -> SubscribeReply:
        ...

    def cancel(self) -> None:
        ...


def _check_valid_response(response: SubscribeReply | None) -> None:
    if not response or response == grpc.aio.EOF:
        msg = "End of stream"
        raise grpc.RpcError(msg)


class EventClient:
    """Generic client class to connect with the Amiga brain services.

    Internally implements an `asyncio` gRPC channel that is designed to be imported
    by service specific clients.
    """

    def __init__(self, config: EventServiceConfig) -> None:
        """Initializes the client.

        Args:
            config (EventServiceConfig): the grpc configuration data structure.

        Raises:
            ValueError: if the port or host are invalid.
        """
        if config.port == 0:
            msg = f"Invalid port: {config}, are you sure this is a client config?"
            raise ValueError(
                msg,
            )

        if config.host == "":
            msg = f"Invalid host: {config}, are you sure this is a client config?"
            raise ValueError(
                msg,
            )

        # the configuration data structure
        self._config: EventServiceConfig = config

        # the client logger
        self._logger: logging.Logger = logging.getLogger(f"{self.config.name}/client")

        # the number of messages sent
        self._metrics = EventServiceHealthMetrics()

        # the gRPC channel and stub
        self.channel: grpc.aio.Channel | None = None
        self.stub: event_service_pb2_grpc.EventServiceStub | None = None

    @property
    def config(self) -> EventServiceConfig:
        """Returns the configuration data structure."""
        return self._config

    @property
    def metrics(self) -> EventServiceHealthMetrics:
        """Returns the metrics data structure."""
        return self._metrics

    @property
    def logger(self) -> logging.Logger:
        """Returns the logger."""
        return self._logger

    @property
    def server_address(self) -> str:
        """Returns the composed address and port."""
        return f"{self.config.host}:{self.config.port}"

    async def _try_connect(self) -> bool:
        """Tries to connect to the server.

        Returns:
            bool: True if the connection was successful, False otherwise.
        """
        if (
            self.stub is not None
            and self.channel is not None
            and await self.channel.channel_ready()
        ):
            self.logger.debug("Already connected to %s", self.server_address)
            return True

        try:
            # create an async connection with the server
            self.channel = grpc.aio.insecure_channel(self.server_address)
            await self.channel.channel_ready()
            self.stub = event_service_pb2_grpc.EventServiceStub(self.channel)

        except OSError as exc:
            self.logger.warning("Could not connect to %s: %s", self.server_address, exc)
            await asyncio.sleep(1)
            return False

        return True

    async def subscribe(
        self,
        request: SubscribeRequest,
        decode: bool = True,
    ) -> AsyncIterator[tuple[Event, Message | bytes]]:
        """Subscribes to the server.

        Args:
            request (SubscribeRequest): the subscription request.
            decode (bool, optional): if True, the payload will be decoded. Defaults to True.

        Yields:
            tuple[Event, Message | bytes]: the event and the payload.
        """
        # the streaming object
        response_stream: EventClientProtocol | None = None

        while True:
            # check if we are connected to the server
            if not await self._try_connect():
                if response_stream is not None:
                    response_stream.cancel()
                    response_stream = None
                self.logger.warning(
                    "%s is not streaming or ready to stream",
                    self.config,
                )
                continue

            if response_stream is None and self.stub is not None:
                # get the streaming object
                response_stream = self.stub.subscribe(request)

            if response_stream is None:
                continue

            try:
                # try/except so app doesn't crash on killed service
                response: SubscribeReply = await response_stream.read()
                response.event.timestamps.append(
                    get_monotonic_now(StampSemantics.CLIENT_RECEIVE),
                )
                _check_valid_response(response)
            except grpc.RpcError as exc:
                self.logger.warning("here %s", exc)
                response_stream.cancel()
                response_stream = None
                continue
            except AssertionError as exc:
                self.logger.warning("End of stream %s", exc)
                response_stream.cancel()
                response_stream = None
                continue

            # decode the payload if requested
            payload_or_protobuf: bytes | Message = response.payload
            if decode:
                payload_or_protobuf = payload_to_protobuf(
                    response.event,
                    response.payload,
                )

            # update the client / server metrics before yielding the response
            self._metrics.update_data(
                f"{response.event.uri.path}/receiver/count",
                self._metrics.get(f"{response.event.uri.path}/receiver/count") + 1,
            )

            stamp_client_receive: float | None = get_stamp_by_semantics_and_clock_type(
                response.event,
                StampSemantics.CLIENT_RECEIVE,
                "monotonic",
            )
            stamp_service_send: float | None = get_stamp_by_semantics_and_clock_type(
                response.event,
                StampSemantics.SERVICE_SEND,
                "monotonic",
            )
            if stamp_client_receive is not None and stamp_service_send is not None:
                self._metrics.update_data(
                    f"{response.event.uri.path}/receiver/latency",
                    stamp_client_receive - stamp_service_send,
                )
            else:
                self.logger.warning(
                    "Could not compute latency: %s %s",
                    stamp_client_receive,
                    stamp_service_send,
                )

            yield response.event, payload_or_protobuf

    async def list_uris(self) -> list[Uri]:
        """Returns the list of uris.

        Returns:
            list[Uri]: the list of uris.
        """
        # try to connect to the server, if it fails return an empty list
        if not await self._try_connect():
            self.logger.warning("Could not list uris: %s", self.server_address)
            return []

        if self.stub is None:
            return []

        try:
            response: ListUrisReply = await self.stub.listUris(ListUrisRequest())
        except grpc.RpcError as err:
            self.logger.warning(
                "Could not list uris: %s\n err=%s",
                self.server_address,
                err,
            )
        else:
            return response.uris
        return []

    async def request_reply(
        self,
        path: str,
        message: Message,
        timestamps: list[Timestamp] | None = None,
        decode: bool = False,
    ) -> RequestReplyReply | Message:
        """Sends a request and waits for a reply.

        Args:
            path (str): the path of the request.
            message (Message): the message to send.
            timestamps (list[Timestamp], optional): the timestamps to add to the event.
                Defaults to None.
            decode (bool, optional): if True, the payload will be decoded. Defaults to False.

        Returns:
            ReqRepReply: the request reply with the event and the payload or the decoded message.
        """
        # try to connect to the server, if it fails return an emmpty response
        if not await self._try_connect():
            self.logger.warning("Could not send: %s", self.server_address)
            return RequestReplyReply()

        # get the current count and increment it
        count: int = int(self._metrics.get(f"{path}/receiver/count"))
        self._metrics.update_data(f"{path}/receiver/count", count + 1)

        # add the timestamps before sending the message
        timestamps = timestamps or []
        timestamps.append(get_monotonic_now(semantics=StampSemantics.CLIENT_SEND))
        timestamps.append(get_system_clock_now(semantics=StampSemantics.CLIENT_SEND))

        # create the request, serialize the message and send it
        uri: Uri = make_proto_uri(path=path, message=message)
        uri.query += f"&service_name={self.config.name}"

        payload = message.SerializeToString()
        request = RequestReplyRequest(
            event=Event(
                uri=uri,
                timestamps=timestamps,
                payload_length=len(payload),
                sequence=count,
            ),
            payload=payload,
        )

        if self.stub is None:
            return RequestReplyReply()

        # send the request and wait for the reply
        reply: RequestReplyReply = await self.stub.requestReply(request)
        reply.event.timestamps.append(
            get_monotonic_now(semantics=StampSemantics.CLIENT_RECEIVE),
        )

        # decode the payload if requested
        reply_or_message: RequestReplyReply | Message = reply
        if decode:
            reply_or_message = payload_to_protobuf(reply.event, reply.payload)

        return reply_or_message


async def test_subscribe(client: EventClient, uri: Uri):
    async for event, payload in client.subscribe(
        SubscribeRequest(uri=uri, every_n=1),
        decode=False,
    ):
        if not isinstance(payload, bytes):
            msg = "payload is not bytes"
            raise TypeError(msg)
        print(client.config.name + event.uri.path, event.sequence, len(payload))


async def test_smoke():
    parser = argparse.ArgumentParser()
    add_service_parser(parser)
    args = parser.parse_args()
    config_list, service_config = load_service_config(args)
    client = EventClient(service_config)
    async_subscriptions: list[asyncio.Task] = []

    uris = []
    while True:
        uris = await client.list_uris()
        if len(uris) > 1:
            break
        await asyncio.sleep(1)

    for uri in uris:
        print(uri.path)
        async_subscriptions.append(asyncio.create_task(test_subscribe(client, uri)))

    await asyncio.gather(*async_subscriptions)


if __name__ == "__main__":

    asyncio.run(test_smoke())



"""
Usage Examples:

Run a test_service from the config:
$ python -m farm_ng.core.event_service --service-config config.json --service-name test_service
"""
from __future__ import annotations

import argparse
import asyncio
import inspect
import logging
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncIterator, Callable

import grpc
from farm_ng.core import event_service_pb2_grpc

# pylint can't find Event or Uri in protobuf generated files
# https://github.com/protocolbuffers/protobuf/issues/10372
from farm_ng.core.event_pb2 import Event
from farm_ng.core.event_service_pb2 import (
    EventServiceConfig,
    EventServiceConfigList,
    ListUrisReply,
    ListUrisRequest,
    RequestReplyReply,
    RequestReplyRequest,
    SubscribeReply,
    SubscribeRequest,
)
from farm_ng.core.events_file_reader import payload_to_protobuf, proto_from_json_file
from farm_ng.core.stamp import (
    StampSemantics,
    get_monotonic_now,
    get_stamp_by_semantics_and_clock_type,
    get_system_clock_now,
)
from farm_ng.core.uri import make_proto_uri
from google.protobuf.empty_pb2 import Empty
from google.protobuf.wrappers_pb2 import Int32Value

if TYPE_CHECKING:
    from farm_ng.core.timestamp_pb2 import Timestamp
    from farm_ng.core.uri_pb2 import Uri
    from google.protobuf.message import Message

from .event_service_metrics import EventServiceHealthMetrics

# public members

__all__ = [
    "EventServiceGrpc",
    "PublishResult",
]


@dataclass(frozen=True)
class PublishResult:
    """The result of publishing a message to the service."""

    # the number of clients the message was sent to
    number_clients: int
    # the sequence number of the message
    sequence_number: int


class EventServiceGrpc:
    """Base class for a gRPC service."""

    # the maximum size of the queue for received messages
    QUEUE_MAX_SIZE: int = 10

    def __init__(
        self,
        server: grpc.aio.Server,
        config: EventServiceConfig,
        logger: logging.Logger | None = None,
    ) -> None:
        """Initializes a new service.

        Args:
            server: The gRPC server instance.
            config: The service configuration.
            logger: The logger instance.
        """
        self._server = server
        self._config = config

        # create a logger for the service

        self._logger: logging.Logger = logger or logging.getLogger(config.name)

        self._logger.setLevel(config.log_level)
        self._logger.addHandler(logging.StreamHandler())

        # the time when the service was started
        self.time_started: float = time.monotonic()

        # the queues for the clients connected to the service
        self._client_queues: dict[
            str,
            list[tuple[SubscribeRequest, asyncio.Queue]],
        ] = {}

        # the URIs of the service
        self._uris: dict[str, Uri] = {}

        # the stats of the URIs, to keep track of dropped messages and other stats
        self._latched_events: dict[str, SubscribeReply] = {}

        self._metrics = EventServiceHealthMetrics()

        # the request/reply handler
        self._request_reply_handler: Callable | None = None

        # NOTE: experimental
        self._decode_request_reply_handler_message: bool = False

        # add the service to the asyncio server
        event_service_pb2_grpc.add_EventServiceServicer_to_server(self, server)

    @property
    def server(self) -> grpc.aio.Server:
        """Returns the server instance."""
        return self._server

    @property
    def config(self) -> EventServiceConfig:
        """Returns the service configuration."""
        return self._config

    @property
    def logger(self) -> logging.Logger:
        """Returns the service logger."""
        return self._logger

    @property
    def metrics(self) -> EventServiceHealthMetrics:
        """Returns the service metrics."""
        return self._metrics

    @property
    def request_reply_handler(self) -> Callable | None:
        """Returns the request/reply handler."""
        return self._request_reply_handler

    @request_reply_handler.setter
    def request_reply_handler(self, handler: Callable) -> None:
        """Sets the request/reply handler."""
        self.add_request_reply_handler(handler)

    def add_request_reply_handler(self, handler: Callable) -> None:
        """Sets the request/reply handler."""

        params = inspect.signature(handler).parameters

        if len(params) not in (1, 2):
            msg = "Request/reply handler must have one or two parameters"
            raise ValueError(
                msg,
            )

        if len(params) == 2:  # noqa: PLR2004
            self._decode_request_reply_handler_message = True

        # is safe to set the handler
        self._request_reply_handler = handler

    @property
    def uris(self) -> dict[str, Uri]:
        """Returns the URIs of the service."""
        return self._uris

    # public methods

    def reset(self) -> None:
        """Resets the service."""
        self._client_queues = {}
        self._uris = {}
        self._latched_events = {}
        self._request_reply_handler = None
        self._metrics.reset_data()

    async def serve(self) -> None:
        """Starts the service.

        This method is called by the service manager.
        """
        self.server.add_insecure_port(f"[::]:{self.config.port}")

        self.logger.info("Starting server on port %i", self.config.port)
        await self.server.start()

        self.logger.info("Server started")

        await asyncio.gather(
            self.server.wait_for_termination(),
            self.publish_health_metrics(),
        )

    async def publish_health_metrics(self) -> None:
        """Publishes the health metrics of the service."""
        while True:
            # publish the health metrics and wait
            await self.publish(path="/health", message=self._metrics.compute_data())
            await asyncio.sleep(1.0)

    async def publish(
        self,
        path: str,
        message: Message,
        timestamps: list[Timestamp] | None = None,
        latch: bool = False,
    ) -> PublishResult:
        """Publish a message to the service.

        Args:
            path (str): The path of the message.
            message (Message): The message.
            timestamps (list[Timestamp], optional): The timestamps. Defaults to None.
        """
        # add the timestamps to the event as it passes through the service
        timestamps = timestamps or []
        timestamps.append(get_monotonic_now(semantics=StampSemantics.SERVICE_SEND))
        timestamps.append(get_system_clock_now(semantics=StampSemantics.SERVICE_SEND))

        # create the URI of the message
        uri: Uri = make_proto_uri(
            path=path,
            message=message,
            service_name=self.config.name,
        )

        # send the message to the service
        result = await self._send_raw(
            uri=uri,
            message=message,
            timestamps=timestamps,
            latch=latch,
        )

        return PublishResult(**result)

    # public gRPC methods

    # pylint: disable=invalid-name
    async def listUris(
        self,
        request: ListUrisRequest,  # pylint: disable=unused-argument
        context: grpc.aio.ServicerContext,  # pylint: disable=unused-argument
    ) -> ListUrisReply:
        """List the URIs of the service.

        Args:
            request (service_pb2.ListUrisRequest): The request.
            context (grpc.aio.ServicerContext): The context.

        Returns:
            service_pb2.ListUrisReply: The reply.
        """
        return ListUrisReply(uris=[uri for _, uri in self._uris.items()])

    async def subscribe(
        self,
        request: SubscribeRequest,
        context: grpc.aio.ServicerContext,
    ) -> AsyncIterator[SubscribeReply]:
        """Implementation of grpc rpc subscribe."""

        if request.uri.path in self._latched_events:
            yield self._latched_events[request.uri.path]
        elif request.uri.path == "*":
            # send all latched items to the special subscriber "*"
            for reply in self._latched_events.values():
                yield reply

        # create a queue for this client
        request_queue: tuple[SubscribeRequest, asyncio.Queue] = (
            request,
            asyncio.Queue(maxsize=self.QUEUE_MAX_SIZE),
        )

        # add the queue to the list of queues
        if request.uri.path not in self._client_queues:
            self._client_queues[request.uri.path] = []

        self._client_queues[request.uri.path].append(request_queue)

        def delete_queue(_) -> None:
            # remove the queue from the list of queues
            self._client_queues[request.uri.path].remove(request_queue)

        # register the callback to delete the queue when the client disconnects
        context.add_done_callback(delete_queue)

        while True:
            yield await request_queue[1].get()

    async def requestReply(
        self,
        request: RequestReplyRequest,
        context: grpc.aio.ServicerContext,  # pylint: disable=unused-argument
    ) -> RequestReplyReply:
        # adds the timestamps to the event as it passes through the service
        recv_stamp = get_monotonic_now(StampSemantics.SERVICE_RECEIVE)
        request.event.timestamps.append(recv_stamp)

        # metadata to return with the reply
        event = Event()
        event.CopyFrom(request.event)
        event.uri.path = "/request" + event.uri.path

        reply_message: Message
        reply_event: Event | None = None

        async def request_handler_wrapper(
            *args,
            **kwargs,
        ) -> tuple[Message, Event | None]:
            if self._request_reply_handler is None:
                self.logger.error("No request/reply handler set, returning Empty()")
                return Empty(), None
            expected_args: list = ["reply_message", "reply_event"]
            result = await self._request_reply_handler(*args, **kwargs)
            if isinstance(result, tuple) and len(result) == len(expected_args):
                return result
            return result, None

        if self._request_reply_handler is not None:
            # decode the requested message to satisfy the handler signature
            if self._decode_request_reply_handler_message:
                message = payload_to_protobuf(request.event, request.payload)
                reply_message, reply_event = await request_handler_wrapper(
                    request.event,
                    message,
                )
            else:
                reply_message, reply_event = await request_handler_wrapper(request)

            if reply_message is None:
                self.logger.error(
                    "Request invalid, please check your request channel and packet %s",
                    request,
                )
        else:
            reply_message = Empty()

        reply_uri: Uri = make_proto_uri(
            path="/reply" + event.uri.path,
            message=reply_message,
        )
        reply_uri.query += f"&service_name={self.config.name}"

        # create the event and send
        reply_payload: bytes = reply_message.SerializeToString()

        timestamps = []
        timestamps.append(get_monotonic_now(semantics=StampSemantics.SERVICE_SEND))
        timestamps.append(get_system_clock_now(semantics=StampSemantics.SERVICE_SEND))

        if reply_event is not None:
            timestamps.extend(
                timestamp
                for timestamp in reply_event.timestamps
                if timestamp.semantics != StampSemantics.SERVICE_SEND
            )

        event = Event(
            uri=reply_uri,
            timestamps=timestamps,
            payload_length=len(reply_payload),
            sequence=event.sequence,
        )

        return RequestReplyReply(event=event, payload=reply_payload)

    # private methods

    async def _publish_event_payload(
        self,
        event: Event,
        payload: bytes,
        latch=False,
    ) -> int:
        """Send an event and payload to the clients.

        Args:
            event (Event): The event.
            payload (bytes): The payload.

        Returns:
            int: The number of clients the message was sent to.
        """
        uri: Uri = event.uri

        # register the URI of the message
        self._uris[uri.path] = uri

        reply = SubscribeReply(
            event=event,
            payload=payload,
        )
        if latch:
            self._latched_events[uri.path] = reply

        # get client queues for the URI
        client_queues: list[tuple[SubscribeRequest, asyncio.Queue]]
        client_queues = self._client_queues.get(uri.path, [])
        client_queues = self._client_queues.get("*", []) + client_queues

        queues: list[asyncio.Queue] = [
            queue
            for request, queue in client_queues
            if request.every_n == 0 or reply.event.sequence % request.every_n == 0
        ]
        self.logger.debug(
            "Sending %s: %d to %d clients",
            uri.path,
            reply.event.sequence,
            len(queues),
        )

        # send the reply to the clients
        for queue in queues:
            try:
                queue.put_nowait(reply)
            except asyncio.QueueFull:
                # keep tracked of dropped messages in a data structure
                # to be able to report them to the user later
                uri_path_dropped: str = f"{uri.path}/dropped"

                num_dropped: int = int(self._metrics.get(uri_path_dropped))
                num_dropped += 1  # increment the count of dropped messages

                # update the metrics with the count of the dropped messages
                self._metrics.update_data(uri_path_dropped, num_dropped)

                self.logger.warning(
                    "dropped %d sequence %d path: %s",
                    num_dropped,
                    reply.event.sequence,
                    uri.path,
                )

        # wait for the queues to be emptied
        await asyncio.sleep(0)

        # return the number of clients the message was sent to
        return len(queues)

    async def _send_raw(
        self,
        uri: Uri,
        message: Message,
        timestamps: list[Timestamp],
        latch: bool,
    ) -> dict[str, int]:
        """Send a message to the service.

        Args:
            uri (Uri): The URI of the message.
            message (Message): The message.
            timestamps (list[Timestamp]): The timestamps.

        Returns:
            dict[str, int]: The sequence number of the message and the number of clients it was sent to.
        """
        # check that the message type is the same as the previous messages
        if uri.path in self._uris:
            # get the URI of the previous message
            previous_uri: Uri = self._uris[uri.path]

            # check that the message type is the same
            if previous_uri.query != uri.query:
                type1: str = previous_uri.query.split("&")[0].split(".")[-1]
                type2: str = uri.query.split("&")[0].split(".")[-1]
                msg = f"Message type mismatch: {type1} != {type2}"
                raise TypeError(msg)

        # register the URI of the message
        self._uris[uri.path] = uri

        # get the count of the URI
        uri_path_send_count: str = f"{uri.path}/send_count"

        count: int = int(self._metrics.get(uri_path_send_count))

        # create the event and send
        payload: bytes = message.SerializeToString()

        event = Event(
            uri=uri,
            timestamps=timestamps,
            payload_length=len(payload),
            sequence=count,
        )

        num_clients: int = await self._publish_event_payload(
            event=event,
            payload=payload,
            latch=latch,
        )

        # update the metrics with the count of the URI
        # NOTE: even if the message is not sent, the count is incremented
        # so that the client can know that the message was not sent by checking
        # the count of the URI.
        self._metrics.update_data(uri_path_send_count, count + 1)

        # update the metrics with the count of the clients the message was sent to
        self._metrics.update_data(f"{uri.path}/send_clients", num_clients)

        # update the metrics with the timestamp of the last message
        # to compute the rate of the messages.
        stamp: float | None = get_stamp_by_semantics_and_clock_type(
            event=event,
            semantics=StampSemantics.SERVICE_SEND,
            clock_type="monotonic",
        )
        if stamp is not None:
            self._metrics.update_stamps(uri.path, stamp)

        return {
            "sequence_number": count,
            "number_clients": num_clients,
        }


def add_service_parser(parser):
    parser.add_argument("--service-config", type=str, required=True)
    parser.add_argument("--service-name", type=str, required=True)


def load_service_config(_args):
    _config_list = proto_from_json_file(_args.service_config, EventServiceConfigList())
    _service_config = None
    for config in _config_list.configs:
        if config.name == _args.service_name:
            _service_config = config
    if _service_config is None:
        msg = f"service {_args.service_name} not found in config list {_config_list}"
        raise ValueError(
            msg,
        )
    return _config_list, _service_config


def event_service_main_args():
    parser = argparse.ArgumentParser()
    add_service_parser(parser)
    args = parser.parse_args()
    return load_service_config(args)


async def test_send_smoke(event_service: EventServiceGrpc) -> None:
    count = 0
    while True:
        await asyncio.sleep(1)
        await event_service.publish("/test", Int32Value(value=count))
        count += 1


async def test_send_smoke2(event_service: EventServiceGrpc) -> None:
    count = 0
    while True:
        await asyncio.sleep(0.5)
        await event_service.publish("/test2", Int32Value(value=count))
        count += 1


async def test_req_rep_handler_smoke(
    request: RequestReplyRequest,
) -> Message:
    message = payload_to_protobuf(request.event, request.payload)
    print(
        f"Received: {request.event.uri.path} {request.event.sequence} {message}".rstrip(),
    )
    return message  # echo message back


# main function to run the service and all the async tasks
async def test_main(args, event_service: EventServiceGrpc) -> None:
    # define the async tasks
    async_tasks = []

    event_service.request_reply_handler = test_req_rep_handler_smoke

    # add the service task
    async_tasks.append(event_service.serve())
    async_tasks.append(test_send_smoke(event_service))
    async_tasks.append(test_send_smoke2(event_service))

    # run the tasks
    await asyncio.gather(*async_tasks)


if __name__ == "__main__":
    config_list: EventServiceConfigList
    service_config: EventServiceConfig
    config_list, service_config = event_service_main_args()
    parser = argparse.ArgumentParser()
    parser.add_argument("--my-arg", type=str, required=True)
    parser.add_argument("foo", type=str, nargs="?")
    args = parser.parse_args(service_config.args)

    # create the gRPC server and initialize the service
    event_service: EventServiceGrpc = EventServiceGrpc(
        grpc.aio.server(),
        service_config,
    )
    loop = asyncio.get_event_loop()
    loop.run_until_complete(test_main(args, event_service))



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from farm_ng.core import event_service_pb2 as farm__ng_dot_core_dot_event__service__pb2

GRPC_GENERATED_VERSION = '1.65.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.66.0'
SCHEDULED_RELEASE_DATE = 'August 6, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in farm_ng/core/event_service_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class EventServiceStub(object):
    """EventService is a generic service for publishing and subscribing to events.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.listUris = channel.unary_unary(
                '/farm_ng.core.proto.EventService/listUris',
                request_serializer=farm__ng_dot_core_dot_event__service__pb2.ListUrisRequest.SerializeToString,
                response_deserializer=farm__ng_dot_core_dot_event__service__pb2.ListUrisReply.FromString,
                _registered_method=True)
        self.subscribe = channel.unary_stream(
                '/farm_ng.core.proto.EventService/subscribe',
                request_serializer=farm__ng_dot_core_dot_event__service__pb2.SubscribeRequest.SerializeToString,
                response_deserializer=farm__ng_dot_core_dot_event__service__pb2.SubscribeReply.FromString,
                _registered_method=True)
        self.requestReply = channel.unary_unary(
                '/farm_ng.core.proto.EventService/requestReply',
                request_serializer=farm__ng_dot_core_dot_event__service__pb2.RequestReplyRequest.SerializeToString,
                response_deserializer=farm__ng_dot_core_dot_event__service__pb2.RequestReplyReply.FromString,
                _registered_method=True)


class EventServiceServicer(object):
    """EventService is a generic service for publishing and subscribing to events.
    """

    def listUris(self, request, context):
        """return a list of all the URIs this service can publish
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def subscribe(self, request, context):
        """subscribe to single URI topic
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def requestReply(self, request, context):
        """send a single message request/reply
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_EventServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'listUris': grpc.unary_unary_rpc_method_handler(
                    servicer.listUris,
                    request_deserializer=farm__ng_dot_core_dot_event__service__pb2.ListUrisRequest.FromString,
                    response_serializer=farm__ng_dot_core_dot_event__service__pb2.ListUrisReply.SerializeToString,
            ),
            'subscribe': grpc.unary_stream_rpc_method_handler(
                    servicer.subscribe,
                    request_deserializer=farm__ng_dot_core_dot_event__service__pb2.SubscribeRequest.FromString,
                    response_serializer=farm__ng_dot_core_dot_event__service__pb2.SubscribeReply.SerializeToString,
            ),
            'requestReply': grpc.unary_unary_rpc_method_handler(
                    servicer.requestReply,
                    request_deserializer=farm__ng_dot_core_dot_event__service__pb2.RequestReplyRequest.FromString,
                    response_serializer=farm__ng_dot_core_dot_event__service__pb2.RequestReplyReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'farm_ng.core.proto.EventService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('farm_ng.core.proto.EventService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class EventService(object):
    """EventService is a generic service for publishing and subscribing to events.
    """

    @staticmethod
    def listUris(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/farm_ng.core.proto.EventService/listUris',
            farm__ng_dot_core_dot_event__service__pb2.ListUrisRequest.SerializeToString,
            farm__ng_dot_core_dot_event__service__pb2.ListUrisReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def subscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/farm_ng.core.proto.EventService/subscribe',
            farm__ng_dot_core_dot_event__service__pb2.SubscribeRequest.SerializeToString,
            farm__ng_dot_core_dot_event__service__pb2.SubscribeReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def requestReply(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/farm_ng.core.proto.EventService/requestReply',
            farm__ng_dot_core_dot_event__service__pb2.RequestReplyRequest.SerializeToString,
            farm__ng_dot_core_dot_event__service__pb2.RequestReplyReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/timestamp.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1c\x66\x61rm_ng/core/timestamp.proto\x12\x12\x66\x61rm_ng.core.proto\"A\n\tTimestamp\x12\r\n\x05stamp\x18\x01 \x01(\x01\x12\x12\n\nclock_name\x18\x02 \x01(\t\x12\x11\n\tsemantics\x18\x03 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.timestamp_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TIMESTAMP']._serialized_start=52
  _globals['_TIMESTAMP']._serialized_end=117
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/linalg.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x66\x61rm_ng/core/linalg.proto\x12\x12\x66\x61rm_ng.core.proto\"\x1f\n\x07Vec2I64\x12\t\n\x01x\x18\x01 \x01(\x03\x12\t\n\x01y\x18\x02 \x01(\x03\"\x1f\n\x07Vec2F32\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\"\x1f\n\x07Vec2F64\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\"*\n\x07Vec3I64\x12\t\n\x01x\x18\x01 \x01(\x03\x12\t\n\x01y\x18\x02 \x01(\x03\x12\t\n\x01z\x18\x03 \x01(\x03\"*\n\x07Vec3F32\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"*\n\x07Vec3F64\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\t\n\x01z\x18\x03 \x01(\x01\"a\n\x07Mat2F32\x12*\n\x05\x63ol_0\x18\x01 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec2F32\x12*\n\x05\x63ol_1\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec2F32\"a\n\x07Mat2F64\x12*\n\x05\x63ol_0\x18\x01 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec2F64\x12*\n\x05\x63ol_1\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec2F64\"\x8d\x01\n\x07Mat3F32\x12*\n\x05\x63ol_0\x18\x01 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F32\x12*\n\x05\x63ol_1\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F32\x12*\n\x05\x63ol_2\x18\x03 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F32\"\x8d\x01\n\x07Mat3F64\x12*\n\x05\x63ol_0\x18\x01 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64\x12*\n\x05\x63ol_1\x18\x02 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64\x12*\n\x05\x63ol_2\x18\x03 \x01(\x0b\x32\x1b.farm_ng.core.proto.Vec3F64\"\x17\n\x07VecXF32\x12\x0c\n\x04\x64\x61ta\x18\x01 \x03(\x02\"\x17\n\x07VecXF64\x12\x0c\n\x04\x64\x61ta\x18\x01 \x03(\x01\"3\n\x07MatXF32\x12\x0c\n\x04rows\x18\x01 \x01(\x05\x12\x0c\n\x04\x63ols\x18\x02 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x03 \x03(\x02\"3\n\x07MatXF64\x12\x0c\n\x04rows\x18\x01 \x01(\x05\x12\x0c\n\x04\x63ols\x18\x02 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x03 \x03(\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.linalg_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_VEC2I64']._serialized_start=49
  _globals['_VEC2I64']._serialized_end=80
  _globals['_VEC2F32']._serialized_start=82
  _globals['_VEC2F32']._serialized_end=113
  _globals['_VEC2F64']._serialized_start=115
  _globals['_VEC2F64']._serialized_end=146
  _globals['_VEC3I64']._serialized_start=148
  _globals['_VEC3I64']._serialized_end=190
  _globals['_VEC3F32']._serialized_start=192
  _globals['_VEC3F32']._serialized_end=234
  _globals['_VEC3F64']._serialized_start=236
  _globals['_VEC3F64']._serialized_end=278
  _globals['_MAT2F32']._serialized_start=280
  _globals['_MAT2F32']._serialized_end=377
  _globals['_MAT2F64']._serialized_start=379
  _globals['_MAT2F64']._serialized_end=476
  _globals['_MAT3F32']._serialized_start=479
  _globals['_MAT3F32']._serialized_end=620
  _globals['_MAT3F64']._serialized_start=623
  _globals['_MAT3F64']._serialized_end=764
  _globals['_VECXF32']._serialized_start=766
  _globals['_VECXF32']._serialized_end=789
  _globals['_VECXF64']._serialized_start=791
  _globals['_VECXF64']._serialized_end=814
  _globals['_MATXF32']._serialized_start=816
  _globals['_MATXF32']._serialized_end=867
  _globals['_MATXF64']._serialized_start=869
  _globals['_MATXF64']._serialized_end=920
# @@protoc_insertion_point(module_scope)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class OptionalG0Float(_message.Message):
    __slots__ = ("value", "has_value")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    HAS_VALUE_FIELD_NUMBER: _ClassVar[int]
    value: float
    has_value: bool
    def __init__(self, value: _Optional[float] = ..., has_value: bool = ...) -> None: ...

class OptionalG0Double(_message.Message):
    __slots__ = ("value", "has_value")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    HAS_VALUE_FIELD_NUMBER: _ClassVar[int]
    value: float
    has_value: bool
    def __init__(self, value: _Optional[float] = ..., has_value: bool = ...) -> None: ...

class OptionalG0Int32(_message.Message):
    __slots__ = ("value", "has_value")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    HAS_VALUE_FIELD_NUMBER: _ClassVar[int]
    value: int
    has_value: bool
    def __init__(self, value: _Optional[int] = ..., has_value: bool = ...) -> None: ...

class OptionalG0Int64(_message.Message):
    __slots__ = ("value", "has_value")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    HAS_VALUE_FIELD_NUMBER: _ClassVar[int]
    value: int
    has_value: bool
    def __init__(self, value: _Optional[int] = ..., has_value: bool = ...) -> None: ...

class OptionalG0Bool(_message.Message):
    __slots__ = ("value", "has_value")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    HAS_VALUE_FIELD_NUMBER: _ClassVar[int]
    value: bool
    has_value: bool
    def __init__(self, value: bool = ..., has_value: bool = ...) -> None: ...

class OptionalG0String(_message.Message):
    __slots__ = ("value", "has_value")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    HAS_VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str
    has_value: bool
    def __init__(self, value: _Optional[str] = ..., has_value: bool = ...) -> None: ...

class FileSystemPath(_message.Message):
    __slots__ = ("path_string",)
    PATH_STRING_FIELD_NUMBER: _ClassVar[int]
    path_string: str
    def __init__(self, path_string: _Optional[str] = ...) -> None: ...
